#!/usr/bin/env python
"""
_ListIncomplete_

MySQL implementation of Subscription.ListIncomplete
"""

from WMCore.Database.DBFormatter import DBFormatter

class ListIncomplete(DBFormatter):
    sql = "SELECT DISTINCT subscription AS id FROM wmbs_sub_files_available"

    def format(self, result):
        results = DBFormatter.format(self, result)

        subIDs = []
        for row in results:
            subIDs.append(row[0])

        return subIDs

    def execute(self, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, conn = conn, transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
# coding=utf-8
"""
Rucio Consistency Monitor Service class developed on top of the native WMCore
Service APIs providing custom output and handling as necessary for the
CMS Workload Management system
"""

from __future__ import division, print_function, absolute_import
from future import standard_library

from urllib.parse import urlencode

import gzip
import json
import logging

from WMCore.Services.Service import Service
from Utils.Utilities import decodeBytesToUnicode


standard_library.install_aliases()


class RucioConMon(Service):
    """
    API for dealing with retrieving information from Rucio Consistency Monitor
    """

    def __init__(self, url, logger=None, configDict=None):
        """
        Init method for the RucioConMon Class
        """
        configDict = configDict or {}
        configDict.setdefault('endpoint', url)
        configDict.setdefault('cacheduration', 1)  # in hours
        configDict.setdefault('accept_type', 'application/json')
        configDict.setdefault('content_type', 'application/json')
        configDict['logger'] = logger if logger else logging.getLogger()
        super(RucioConMon, self).__init__(configDict)
        self['logger'].debug("Initializing RucioConMon with url: %s", self['endpoint'])

    def _getResult(self, uri, callname="", clearCache=False, args=None, binary=False):
        """
        Either fetch data from the cache file or query the data-service
        :param uri: The endpoint uri
        :param callname: alias for caller function
        :param clearCache: parameter to control the cache behavior
        :param args: additional parameters to HTTP request call
        :param binary: specifies request for binary object from HTTP requests (e.g. zipped content)
        :return:    A dictionary
        """

        # NOTE: Unlike the common case we are not using the callname for building
        #       apiUrl. we  are using it only for giving proper name of the cache
        #       file.
        # NOTE: The 'callname' should not contain '/', otherwise the cache is
        #       tried to be created with a a non existing subdirectory structure

        cachedApi = "%s.json" % callname
        # apiUrl = '%s?json&preset=%s' % (uri, callname)
        apiUrl = uri

        self['logger'].debug('Fetching data from %s, with args %s', apiUrl, args)
        if args:
            apiUrl = "%s&%s" % (apiUrl, urlencode(args, doseq=True))

        if clearCache:
            self.clearCache(cachedApi, args)
        results = '{}' # explicitly define results which will be loaded by json.loads below
        if binary:
            with self.refreshCache(cachedApi, apiUrl, decoder=False, binary=True) as istream:
                results = gzip.decompress(istream.read())
            return results
        else:
            with self.refreshCache(cachedApi, apiUrl) as istream:
                results = istream.read()

        results = json.loads(results)
        return results

    def _getResultZipped(self, uri, callname="", clearCache=True, args=None):
        """
        This method is retrieving a zipped file from the uri privided, instead
        of the normal json
        :param uri: The endpoint uri
        :param callname: alias for caller function
        :param clearCache: parameter to control the cache behavior
        :param args: additional parameters to HTTP request call
        :return:    a list of LFNs
        """
        data = self._getResult(uri, callname, clearCache, args, binary=True)
        # convert bytes which we received upstream to string
        data = decodeBytesToUnicode(data)
        return [f for f in data.split('\n') if f]

    def getRSEStats(self):
        """
        Gets the latest statistics from the RucioConMon, together with the last
        update timestamps for all RSEs known to CMS Rucio
        :return: A dictionary
        """
        uri = "stats"
        rseStats = self._getResult(uri, callname='stats')
        return rseStats

    def getRSEUnmerged(self, rseName, zipped=False):
        """
        Gets the list of all unmerged files in an RSE
        :param rseName: The RSE whose list of unmerged files to be retrieved
        :param zipped:  If True the interface providing the zipped lists will be called
        :return:        A list of unmerged files for the RSE in question
        """
        # NOTE: The default API provided by Rucio Consistency Monitor is in a form of a
        #       zipped file/stream. Currently we are using the newly provided json API
        #       But in in case we figure out the data is too big we may need to
        #       implement the method with the zipped API and use disc cache for
        #       reading/streaming from file. This will prevent any set arithmetic
        #       in the future.
        if not zipped:
            uri = "files?rse=%s&format=json" % rseName
            rseUnmerged = self._getResult(uri, callname=rseName)
            return rseUnmerged
        else:
            uri = "files?rse=%s&format=raw" % rseName
            callname = '{}.zipped'.format(rseName)
            rseUnmerged = self._getResultZipped(uri,  callname=callname, clearCache=True)
            return rseUnmerged



#!/usr/bin/env python
"""
_SetDatasetAlgo_

MySQL implementation of DBSUpload.SetDatabaseAlgo
Should set the database-algo inDBS switch
"""




from WMCore.Database.DBFormatter import DBFormatter

class SetDatasetAlgo(DBFormatter):
    """
    Set the inDBS status of a datasetAlgo

    """

    sql = """UPDATE dbsbuffer_algo_dataset_assoc das
             SET das.in_dbs = :in_dbs
             WHERE das.id = :datasetAlgo
    """

    def execute(self, datasetAlgo, inDBS = 1, conn = None, transaction = False):
        """
        _execute_

        Set the datasetAlgo
        """
        binds = {"datasetAlgo": datasetAlgo, 'in_dbs': inDBS}

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_ChangeState_

Propagate a job from one state to another.
"""
import json
import sys
from builtins import str
import logging
import re
import time

from WMCore.DataStructs.WMObject import WMObject
from WMCore.Database.CMSCouch import CouchNotFoundError, CouchError, CouchRequestTooLargeError
from WMCore.Database.CMSCouch import CouchServer
from WMCore.JobStateMachine.SummaryDB import updateSummaryDB
from WMCore.JobStateMachine.Transitions import Transitions
from WMCore.Lexicon import sanitizeURL
from WMCore.WMConnectionBase import WMConnectionBase
from WMCore.WMSpec.WMWorkload import WMWorkloadHelper

CMSSTEP = re.compile(r'^cmsRun[0-9]+$')


def discardConflictingDocument(couchDbInstance, data, result):
    """
    _discardConflictingDocument_

    This should be passed to the queue and commit calls of CMSCouch
    in order to tell it what to do with conflicting documents.
    In this case we trash the old one and replace with what we were
    trying to commit, this is available in the data argument.
    And the result tells us the id of the conflicting document
    """

    conflictingId = result["id"]

    try:
        if not couchDbInstance.documentExists(conflictingId):
            # It doesn't exist, this is odd
            # Don't try again
            return result

        # Get the revision to override
        originalDocRev = couchDbInstance.document(conflictingId)["_rev"]

        # Look for the data to be commited
        retval = result
        for doc in data["docs"]:
            if doc["_id"] == conflictingId:
                doc["_rev"] = originalDocRev
                retval = couchDbInstance.commitOne(doc)
                break

        return retval
    except CouchError as ex:
        logging.error("Couldn't resolve conflict when updating document with id %s", result["id"])
        logging.error("Error: %s", str(ex))
        return result


def shrinkLargeFJR(couchDbInstance, sizeLimit):
    """
    Look at the CouchDB database queue and empty documents
    with a 'fwjr' field larger than sizeLimit.
    :param couchDbInstance: couchdb database instance
    :param sizeLimit: integer with the limit number of bytes
    :return: None
    """
    for doc in couchDbInstance._queue:
        if sys.getsizeof(json.dumps(doc)) > sizeLimit:
            if 'fwjr' in doc:
                errMsg = f"The 'fwjr' attribute for jobid: {doc.get('jobid')} will be empty "
                errMsg += f"because it is larger than the configured limit: {sizeLimit} bytes."
                logging.warning(errMsg)
                doc['fwjr'] = dict()
            else:
                errMsg = f"Found a too large Couch document for job id: {doc.get('jobid')}, which is "
                errMsg += f"actually missing the 'fwjr' attribute. Current limit: {sizeLimit} bytes."
                logging.error(errMsg)
    return


def getDataFromSpecFile(specFile):
    workload = WMWorkloadHelper()
    workload.load(specFile)
    campaign = workload.getCampaign()
    result = {"Campaign": campaign}
    for task in workload.taskIterator():
        result[task.getPathName()] = task.getPrepID()
    return result


class ChangeState(WMObject, WMConnectionBase):
    """
    Propagate the state of a job through the JSM.
    """

    def __init__(self, config, couchDbName=None):
        WMObject.__init__(self, config)
        WMConnectionBase.__init__(self, "WMCore.WMBS")

        if couchDbName is None:
            self.dbname = getattr(self.config.JobStateMachine, "couchDBName")
        else:
            self.dbname = couchDbName

        self.jobsdatabase = None
        self.fwjrdatabase = None
        self.jsumdatabase = None
        self.statsumdatabase = None

        # max total number of documents to be committed in the same Couch operation
        self.maxBulkCommit = getattr(self.config.JobStateMachine, 'maxBulkCommitDocs', 250)
        self.couchdb = CouchServer(self.config.JobStateMachine.couchurl)
        self._connectDatabases()

        self.getCouchDAO = self.daofactory("Jobs.GetCouchID")
        self.setCouchDAO = self.daofactory("Jobs.SetCouchID")
        self.incrementRetryDAO = self.daofactory("Jobs.IncrementRetry")
        self.workflowTaskDAO = self.daofactory("Jobs.GetWorkflowTask")
        self.jobTypeDAO = self.daofactory("Jobs.GetType")
        self.updateLocationDAO = self.daofactory("Jobs.UpdateLocation")
        self.getWorkflowSpecDAO = self.daofactory("Workflow.GetSpecAndNameFromTask")

        self.maxUploadedInputFiles = getattr(self.config.JobStateMachine, 'maxFWJRInputFiles', 1000)
        self.fwjrLimitSize = getattr(self.config.JobStateMachine, 'fwjrLimitSize', 8 * 1000**2)
        self.workloadCache = {}
        return

    def _connectDatabases(self):
        """
        Try connecting to the couchdbs
        """
        if not hasattr(self, 'jobsdatabase') or self.jobsdatabase is None:
            try:
                self.jobsdatabase = self.couchdb.connectDatabase("%s/jobs" % self.dbname,
                                                                 size=self.maxBulkCommit)
            except Exception as ex:
                logging.error("Error connecting to couch db '%s/jobs': %s", self.dbname, str(ex))
                self.jobsdatabase = None
                return False

        if not hasattr(self, 'fwjrdatabase') or self.fwjrdatabase is None:
            try:
                self.fwjrdatabase = self.couchdb.connectDatabase("%s/fwjrs" % self.dbname,
                                                                 size=self.maxBulkCommit)
            except Exception as ex:
                logging.error("Error connecting to couch db '%s/fwjrs': %s", self.dbname, str(ex))
                self.fwjrdatabase = None
                return False

        if not hasattr(self, 'jsumdatabase') or self.jsumdatabase is None:
            dbname = getattr(self.config.JobStateMachine, 'jobSummaryDBName')
            try:
                self.jsumdatabase = self.couchdb.connectDatabase(dbname,
                                                                 size=self.maxBulkCommit)
            except Exception as ex:
                logging.error("Error connecting to couch db '%s': %s", dbname, str(ex))
                self.jsumdatabase = None
                return False

        if not hasattr(self, 'statsumdatabase') or self.statsumdatabase is None:
            dbname = getattr(self.config.JobStateMachine, 'summaryStatsDBName')
            try:
                self.statsumdatabase = self.couchdb.connectDatabase(dbname,
                                                                    size=self.maxBulkCommit)
            except Exception as ex:
                logging.error("Error connecting to couch db '%s': %s", dbname, str(ex))
                self.jsumdatabase = None
                return False

        return True

    def propagate(self, jobs, newstate, oldstate, updatesummary=False):
        """
        Move the job from a state to another. Book keep the change to CouchDB.
        Report the information to the Dashboard.
        Take a list of job objects (dicts) and the desired state change.
        Return the jobs back, throw assertion error if the state change is not allowed
        and other exceptions as appropriate
        """
        if not isinstance(jobs, list):
            jobs = [jobs]

        if len(jobs) == 0:
            return

        # 1. Is the state transition allowed?
        self.check(newstate, oldstate)

        # 2. Load workflow/task information into the jobs
        self.loadExtraJobInformation(jobs)

        # 3. Make the state transition
        self.persist(jobs, newstate, oldstate)

        # 4. Complete the job information for jobs in created state
        try:
            self.completeCreatedJobsInformation(jobs, newstate, oldstate)
        except Exception as ex:
            logging.exception("Error complementing created job information: %s", str(ex))

        # 5. Document the state transition in couch
        try:
            self.recordInCouch(jobs, newstate, oldstate, updatesummary)
        except UnicodeDecodeError as ex:
            msg = "A critical error happened! Report it to developers. Error: %s" % str(ex)
            logging.exception(msg)
            raise
        except Exception as ex:
            logging.exception("Error updating job in couch: %s", str(ex))

        return

    def check(self, newstate, oldstate):
        """
        check that the transition is allowed. return a tuple of the transition
        if it is allowed, throw up an exception if not.
        """
        newstate = newstate.lower()
        oldstate = oldstate.lower()

        # Check for wrong transitions
        transitions = Transitions()
        assert newstate in transitions[oldstate], \
            "Illegal state transition requested: %s -> %s" % (oldstate, newstate)

    def recordInCouch(self, jobs, newstate, oldstate, updatesummary=False):
        """
        _recordInCouch_

        Record relevant job information in couch. If the job does not yet exist
        in couch it will be saved as a seperate document.  If the job has a FWJR
        attached that will be saved as a seperate document.
        """
        if not self._connectDatabases():
            logging.error('Databases not connected properly')
            return

        timestamp = int(time.time())
        couchRecordsToUpdate = []

        countDocs = 0
        for job in jobs:
            couchDocID = job.get("couch_record", None)

            if newstate == "new":
                oldstate = "none"

            if job.get("site_cms_name", None):
                if newstate == "executing":
                    jobLocation = job["site_cms_name"]
                else:
                    jobLocation = "Agent"
            else:
                jobLocation = "Agent"

            if couchDocID is None:
                jobDocument = {}
                jobDocument["_id"] = str(job["id"])
                job["couch_record"] = jobDocument["_id"]
                jobDocument["jobid"] = job["id"]
                jobDocument["workflow"] = job["workflow"]
                jobDocument["task"] = job["task"]
                jobDocument["owner"] = job["owner"]

                jobDocument["inputfiles"] = []
                for inputFile in job["input_files"]:
                    docInputFile = inputFile.json()

                    docInputFile["parents"] = []
                    for parent in inputFile["parents"]:
                        docInputFile["parents"].append({"lfn": parent["lfn"]})

                    jobDocument["inputfiles"].append(docInputFile)

                jobDocument["states"] = {"0": {"oldstate": oldstate,
                                               "newstate": newstate,
                                               "location": jobLocation,
                                               "timestamp": timestamp}}

                jobDocument["jobgroup"] = job["jobgroup"]
                jobDocument["mask"] = {"FirstEvent": job["mask"]["FirstEvent"],
                                       "LastEvent": job["mask"]["LastEvent"],
                                       "FirstLumi": job["mask"]["FirstLumi"],
                                       "LastLumi": job["mask"]["LastLumi"],
                                       "FirstRun": job["mask"]["FirstRun"],
                                       "LastRun": job["mask"]["LastRun"]}

                if job['mask']['runAndLumis'] != {}:
                    # Then we have to save the mask runAndLumis
                    jobDocument['mask']['runAndLumis'] = {}
                    for key in job['mask']['runAndLumis']:
                        jobDocument['mask']['runAndLumis'][str(key)] = job['mask']['runAndLumis'][key]

                jobDocument["name"] = job["name"]
                jobDocument["type"] = "job"
                jobDocument["user"] = job.get("user", None)
                jobDocument["group"] = job.get("group", None)
                jobDocument["taskType"] = job.get("taskType", "Unknown")
                jobDocument["jobType"] = job.get("jobType", "Unknown")

                couchRecordsToUpdate.append({"jobid": job["id"],
                                             "couchid": jobDocument["_id"]})
                if countDocs >= self.jobsdatabase.getQueueSize():
                    self.jobsdatabase.commit(callback=discardConflictingDocument)
                self.jobsdatabase.queue(jobDocument, callback=discardConflictingDocument)
            else:
                # We send a PUT request to the stateTransition update handler.
                # Couch expects the parameters to be passed as arguments to in
                # the URI while the Requests class will only encode arguments
                # this way for GET requests.  Changing the Requests class to
                # encode PUT arguments as couch expects broke a bunch of code so
                # we'll just do our own encoding here.
                updateUri = "/" + self.jobsdatabase.name + "/_design/JobDump/_update/stateTransition/" + couchDocID
                updateUri += "?oldstate=%s&newstate=%s&location=%s&timestamp=%s" % (oldstate,
                                                                                    newstate,
                                                                                    jobLocation,
                                                                                    timestamp)
                self.jobsdatabase.makeRequest(uri=updateUri, type="PUT", decode=False)

            # updating the status of the summary doc only when it is explicitely requested
            # doc is already in couch
            if updatesummary:
                jobSummaryId = job["name"]
                updateUri = "/" + self.jsumdatabase.name + "/_design/WMStatsAgent/_update/jobSummaryState/" + jobSummaryId
                # map retrydone state to jobfailed state for monitoring
                if newstate == "retrydone":
                    monitorState = "jobfailed"
                else:
                    monitorState = newstate
                updateUri += "?newstate=%s&timestamp=%s" % (monitorState, timestamp)
                self.jsumdatabase.makeRequest(uri=updateUri, type="PUT", decode=False)
                logging.debug("Updated job summary status for job %s", jobSummaryId)

                updateUri = "/" + self.jsumdatabase.name + "/_design/WMStatsAgent/_update/jobStateTransition/" + jobSummaryId
                updateUri += "?oldstate=%s&newstate=%s&location=%s&timestamp=%s" % (oldstate,
                                                                                    monitorState,
                                                                                    job["location"],
                                                                                    timestamp)
                self.jsumdatabase.makeRequest(uri=updateUri, type="PUT", decode=False)
                logging.debug("Updated job summary state history for job %s", jobSummaryId)

            if job.get("fwjr", None):

                cachedByWorkflow = self.workloadCache.setdefault(job['workflow'],
                                                                 getDataFromSpecFile(
                                                                     self.getWorkflowSpecDAO.execute(job['task'])[
                                                                         job['task']]['spec']))
                job['fwjr'].setCampaign(cachedByWorkflow.get('Campaign', ''))
                job['fwjr'].setPrepID(cachedByWorkflow.get(job['task'], ''))
                # If there are too many input files, strip them out
                # of the FWJR, as they should already
                # be in the database
                # This is not critical
                try:
                    if len(job['fwjr'].getAllInputFiles()) > self.maxUploadedInputFiles:
                        job['fwjr'].stripInputFiles()
                except Exception as ex:
                    logging.error("Error while trying to strip input files from FWJR.  Ignoring. : %s", str(ex))

                if newstate == "retrydone":
                    jobState = "jobfailed"
                else:
                    jobState = newstate

                # there is race condition updating couch record location and job is completed.
                # for the fast fail job, it could miss the location update
                job["location"] = job["fwjr"].getSiteName() or job.get("location", "Unknown")
                # complete fwjr document
                job["fwjr"].setTaskName(job["task"])
                jsonFWJR = job["fwjr"].__to_json__(None)

                # Don't archive cleanup job report
                if job["jobType"] == "Cleanup":
                    archStatus = "skip"
                else:
                    archStatus = "ready"

                fwjrDocument = {"_id": "%s-%s" % (job["id"], job["retry_count"]),
                                "jobid": job["id"],
                                "jobtype": job["jobType"],
                                "jobstate": jobState,
                                "retrycount": job["retry_count"],
                                "archivestatus": archStatus,
                                "fwjr": jsonFWJR,
                                "type": "fwjr"}
                if countDocs >= self.fwjrdatabase.getQueueSize():
                    try:
                        # TODO: CouchDB bulk insert may fail at any given document. That means, only by
                        # parsing the response object we can actually see which documents succeeded or not.
                        # https://docs.couchdb.org/en/stable/api/database/bulk-api.html#updating-documents-in-bulk
                        self.fwjrdatabase.commit(callback=discardConflictingDocument)
                    except CouchRequestTooLargeError as exc:
                        msg = "Failed to commit bulk of framework job report to CouchDB."
                        msg += f" Details: {str(exc)}"
                        logging.warning(msg)
                        shrinkLargeFJR(self.fwjrdatabase, self.fwjrLimitSize)
                        # now all the documents should fit in
                        self.fwjrdatabase.commit(callback=discardConflictingDocument)

                self.fwjrdatabase.queue(fwjrDocument, timestamp=True, callback=discardConflictingDocument)

                updateSummaryDB(self.statsumdatabase, job)

                # TODO: can add config switch to swich on and off
                # if self.config.JobSateMachine.propagateSuccessJobs or (job["retry_count"] > 0) or (newstate != 'success'):
                if (job["retry_count"] > 0) or (newstate != 'success'):
                    jobSummaryId = job["name"]
                    # building a summary of fwjr
                    logging.debug("Pushing job summary for job %s", jobSummaryId)
                    errmsgs = {}
                    inputs = []
                    if "steps" in fwjrDocument["fwjr"]:
                        for step in fwjrDocument["fwjr"]["steps"]:
                            if "errors" in fwjrDocument["fwjr"]["steps"][step]:
                                errmsgs[step] = [error for error in fwjrDocument["fwjr"]["steps"][step]["errors"]]
                            if "input" in fwjrDocument["fwjr"]["steps"][step] and "source" in \
                                    fwjrDocument["fwjr"]["steps"][step]["input"]:
                                inputs.extend(
                                    [source["runs"] for source in fwjrDocument["fwjr"]['steps'][step]["input"]["source"]
                                     if "runs" in source])

                    outputs = []
                    outputDataset = None
                    for singlestep in job["fwjr"].listSteps():
                        for singlefile in job["fwjr"].getAllFilesFromStep(step=singlestep):
                            if singlefile:
                                if len(singlefile.get('locations', set())) > 1:
                                    locations = list(singlefile.get('locations'))
                                elif singlefile.get('locations'):
                                    locations = singlefile['locations'].pop()
                                else:
                                    locations = set()
                                if CMSSTEP.match(singlestep):
                                    outType = 'output'
                                else:
                                    outType = singlefile.get('module_label', None)
                                outputs.append({'type': outType,
                                                'lfn': singlefile.get('lfn', None),
                                                'location': locations,
                                                'checksums': singlefile.get('checksums', {}),
                                                'size': singlefile.get('size', None)})
                                # it should have one output dataset for all the files
                                outputDataset = singlefile.get('dataset', None) if not outputDataset else outputDataset
                    inputFiles = []
                    for inputFileStruct in job["fwjr"].getAllInputFiles():
                        # check if inputFileSummary needs to be extended
                        inputFileSummary = {}
                        inputFileSummary["lfn"] = inputFileStruct["lfn"]
                        inputFileSummary["input_type"] = inputFileStruct["input_type"]
                        inputFiles.append(inputFileSummary)

                    # Don't record intermediate jobfailed status in the jobsummary
                    # change to jobcooloff which will be overwritten by error handler anyway
                    if (job["retry_count"] > 0) and (newstate == 'jobfailed'):
                        summarystate = 'jobcooloff'
                    else:
                        summarystate = newstate

                    jobSummary = {"_id": jobSummaryId,
                                  "wmbsid": job["id"],
                                  "type": "jobsummary",
                                  "retrycount": job["retry_count"],
                                  "workflow": job["workflow"],
                                  "task": job["task"],
                                  "jobtype": job["jobType"],
                                  "state": summarystate,
                                  "site": job.get("location", None),
                                  "cms_location": job["fwjr"].getSiteName(),
                                  "exitcode": job["fwjr"].getExitCode(),
                                  "eos_log_url": job["fwjr"].getLogURL(),
                                  "worker_node_info": job["fwjr"].getWorkerNodeInfo(),
                                  "errors": errmsgs,
                                  "lumis": inputs,
                                  "outputdataset": outputDataset,
                                  "inputfiles": inputFiles,
                                  "acdc_url": "%s/%s" % (
                                  sanitizeURL(self.config.ACDC.couchurl)['url'], self.config.ACDC.database),
                                  "agent_name": self.config.Agent.hostName,
                                  "output": outputs}
                    if couchDocID is not None:
                        try:
                            currentJobDoc = self.jsumdatabase.document(id=jobSummaryId)
                            jobSummary['_rev'] = currentJobDoc['_rev']
                            jobSummary['state_history'] = currentJobDoc.get('state_history', [])
                            # record final status transition
                            if newstate == 'success':
                                finalStateDict = {'oldstate': oldstate,
                                                  'newstate': newstate,
                                                  'location': job["location"],
                                                  'timestamp': timestamp}
                                jobSummary['state_history'].append(finalStateDict)

                            noEmptyList = ["inputfiles", "lumis"]
                            for prop in noEmptyList:
                                jobSummary[prop] = jobSummary[prop] if jobSummary[prop] else currentJobDoc.get(prop, [])
                        except CouchNotFoundError:
                            pass
                    if countDocs >= self.fwjrdatabase.getQueueSize():
                        self.jsumdatabase.commit()
                    self.jsumdatabase.queue(jobSummary, timestamp=True)

        if len(couchRecordsToUpdate) > 0:
            self.setCouchDAO.execute(bulkList=couchRecordsToUpdate,
                                     conn=self.getDBConn(),
                                     transaction=self.existingTransaction())

        try:
            self.fwjrdatabase.commit(callback=discardConflictingDocument)
        except CouchRequestTooLargeError as exc:
            msg = "Failed to commit bulk of framework job report to CouchDB."
            msg += f" Error: {str(exc)}"
            logging.warning(msg)
            shrinkLargeFJR(self.fwjrdatabase, self.fwjrLimitSize)
            # now all the documents should fit in
            self.fwjrdatabase.commit(callback=discardConflictingDocument)
        self.jobsdatabase.commit(callback=discardConflictingDocument)
        self.jsumdatabase.commit()
        return

    def persist(self, jobs, newstate, oldstate):
        """
        _persist_

        Update the job state in the database.
        """

        if newstate == "killed":
            self.incrementRetryDAO.execute(jobs, increment=99999,
                                           conn=self.getDBConn(),
                                           transaction=self.existingTransaction())
        elif oldstate == "submitcooloff" or oldstate == "jobcooloff" or oldstate == "createcooloff":
            self.incrementRetryDAO.execute(jobs,
                                           conn=self.getDBConn(),
                                           transaction=self.existingTransaction())
        for job in jobs:
            job['state'] = newstate
            job['oldstate'] = oldstate

        dao = self.daofactory(classname="Jobs.ChangeState")
        dao.execute(jobs, conn=self.getDBConn(),
                    transaction=self.existingTransaction())

    def loadExtraJobInformation(self, jobs):
        # This is needed for both couch and dashboard
        jobIDsToCheck = []
        jobTasksToCheck = []
        # This is for mapping ids to the position in the list
        jobMap = {}
        for idx, job in enumerate(jobs):
            if job["couch_record"] is None:
                jobIDsToCheck.append(job["id"])
            if job.get("task", None) is None or job.get("workflow", None) is None \
                    or job.get("taskType", None) is None or job.get("jobType", None) is None:
                jobTasksToCheck.append(job["id"])
            jobMap[job["id"]] = idx

        if len(jobIDsToCheck) > 0:
            couchIDs = self.getCouchDAO.execute(jobID=jobIDsToCheck,
                                                conn=self.getDBConn(),
                                                transaction=self.existingTransaction())
            for couchID in couchIDs:
                idx = jobMap[couchID["jobid"]]
                jobs[idx]["couch_record"] = couchID["couch_record"]
        if len(jobTasksToCheck) > 0:
            jobTasks = self.workflowTaskDAO.execute(jobIDs=jobTasksToCheck,
                                                    conn=self.getDBConn(),
                                                    transaction=self.existingTransaction())
            for jobTask in jobTasks:
                idx = jobMap[jobTask["id"]]
                jobs[idx]["task"] = jobTask["task"]
                jobs[idx]["workflow"] = jobTask["name"]
                jobs[idx]["taskType"] = jobTask["type"]
                jobs[idx]["jobType"] = jobTask["subtype"]

    def completeCreatedJobsInformation(self, jobs, newstate, oldstate):
        """
        This method adds some extra information to the job object, if its
        new state is 'created' and its was in 'cooloff'.
        This is required for jobs coming from ErrorHandler, RetryManager
        or from the T0 pause script, where standard information of a
        WMBSJob is provided.
        :param jobs: list of job objects (dictionaries)
        :param newstate: string with the new state for this set of jobs
        :param oldstate: string with the previous state for this set of jobs
        :return: updates the job objects in-place
        """
        if newstate != 'created':
            return

        incrementRetry = True if 'cooloff' in oldstate else False
        for job in jobs:
            # It there's no jobID in the mask then it's not loaded
            if "jobID" not in job["mask"]:
                # Make sure the daofactory was not stripped
                if getattr(job["mask"], "daofactory", None):
                    job["mask"].load(jobID=job["id"])
            # If the mask is event based, then we have info to report
            if job["mask"]['inclusivemask'] and job["mask"]["LastEvent"] is not None and \
                            job["mask"]["FirstEvent"] is not None:
                job["nEventsToProc"] = int(job["mask"]["LastEvent"] -
                                           job["mask"]["FirstEvent"] + 1)
            # Increment retry when commanded
            if incrementRetry:
                job["retry_count"] += 1

    def recordLocationChange(self, jobs):
        """
        _recordLocationChange_

        Record a location change in couch and WMBS,
        this expects a list of dictionaries with
        jobid and location keys which represent
        the job id in WMBS and new location respectively.
        """
        if not self._connectDatabases():
            logging.error('Databases not connected properly')
            return

        # First update safely in WMBS
        self.updateLocationDAO.execute(jobs, conn=self.getDBConn(),
                                       transaction=self.existingTransaction())
        # Now try couch, this can fail and we don't require it to succeed
        try:
            jobIDs = [x['jobid'] for x in jobs]
            couchIDs = self.getCouchDAO.execute(jobIDs, conn=self.getDBConn(),
                                                transaction=self.existingTransaction())
            locationCache = dict((x['jobid'], x['location']) for x in jobs)
            for entry in couchIDs:
                couchRecord = entry['couch_record']
                location = locationCache[entry['jobid']]
                updateUri = "/" + self.jobsdatabase.name + "/_design/JobDump/_update/locationTransition/" + couchRecord
                updateUri += "?location=%s" % (location)
                self.jobsdatabase.makeRequest(uri=updateUri, type="PUT", decode=False)
        except Exception as ex:
            logging.exception("Error updating job in couch: %s", str(ex))



#!/bin/env python
"""
_PauseAlgo_
"""
import os.path
import logging
from WMComponent.RetryManager.PlugIns.RetryAlgoBase import RetryAlgoBase
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.FwkJobReport.Report import Report


class PauseAlgo(RetryAlgoBase):
    """
    _PauseAlgo_

        This implements the Paused job algorithm, explanation of the concept in #3114
    """

    def __init__(self, config):
        RetryAlgoBase.__init__(self, config)
        self.changer = ChangeState(config)

    def isReady(self, job, cooloffType):
        """
        Actual function that does the work
        """
        # This should come from configuration, pause_count

        pauseCount = self.getAlgoParam(job['jobType'], param='pauseCount', defaultReturn=3)

        pauseMap = {
            'createcooloff': 'createpaused',
            'submitcooloff': 'submitpaused',
            'jobcooloff': 'jobpaused'
            }

        # Setting a pauseCount depending on job exit code
        if job['state'] == 'jobcooloff':
            exitCodes = self.getAlgoParam(job['jobType'], 'retryErrorCodes', {})
            if exitCodes:
                report = Report()
                reportPath = os.path.join(job['cache_dir'], "Report.%i.pkl" % job['retry_count'])
                try:
                    report.load(reportPath)
                    jobExitCode = report.getExitCode()
                    # If the jobExitCode is configured, set the respective pauseCount for the job.
                    if jobExitCode in exitCodes:
                        retryByTimeOut = True
                        pauseCount = exitCodes[jobExitCode]
                except IOError as ex:
                    msg = "Error loading report %s\n" % (reportPath)
                    msg += str(ex)
                    logging.warning(msg)

        # Here introduces the SquaredAlgo logic :
        baseTimeoutDict = self.getAlgoParam(job['jobType'])
        baseTimeout = baseTimeoutDict.get(cooloffType.lower(), 10)
        cooloffTime = baseTimeout * pow(job['retry_count'], 2)
        currentTime = self.timestamp()
        if currentTime - job['state_time'] > cooloffTime:
            retryByTimeOut = True
        else:
            retryByTimeOut = False

        if retryByTimeOut:
            # If reached the pauseCount, we want the job to pause instead of retrying
            if pauseCount == 0:
                self.changer.propagate(job, pauseMap[job['state']], job['state'], updatesummary=True)
                return False
            elif job['retry_count'] > 0 and not job['retry_count'] % pauseCount:
                self.changer.propagate(job, pauseMap[job['state']], job['state'], updatesummary=True)
                return False
            else:
                return True
        else:
            return False



#!/usr/bin/env python
"""
_StageOutMgrV2_

Refactoring of StageOutMgr -- for now, not accessed by default


"""

from builtins import range, object

import logging
import time
import traceback

# PyCharm likes to remove these two imports, but we cannot let it do it.
import WMCore.Storage.Backends
import WMCore.Storage.Plugins

from WMCore.Storage.Registry import RegistryError, retrieveStageOutImpl
from WMCore.Storage.SiteLocalConfig import loadSiteLocalConfig
from WMCore.Storage.StageOutError import StageOutError, StageOutFailure

log = logging


class FileManager(object):
    """
    _FileManager_

    Object that handles modifying files in a site-specific way.
    Supercedes StageInMgr, StageOutMgr, DeleteMgr

    new easy to use interface:
    deleteLFN(lfn) - tries to delete a certain LFN, returning details on success. Raising on failure
    stageIn/stageOut - accepts a dict containing the details
                        PFN stores the local file name
                        LFN stores the lfn of the file, which will be mapped to a PFN

    plugin implementations require:
    newPfn =  pluginImplementation.doTransfer( lfn, pfn, stageOut, pnn, command, options, protocol  )
    pluginImplementation.doDelete(pfn, pnn, command, options, protocol  )

    Which make one attempt to perform the action and raises if it doesn't succeed. It is the plugins
    responsibility to verify that things are complete.
    """

    def __init__(self, numberOfRetries=3, retryPauseTime=15, **overrideParams):

        # set defaults
        self.firstException = None
        self.failed = {}
        self.completedFiles = {}
        self.override = False
        self.overrideConf = overrideParams
        self.substituteGUID = True
        self.defaultMethod = {}
        self.fallbacks = []
        self.tfc = None
        self.numberOfRetries = numberOfRetries
        self.retryPauseTime = retryPauseTime

        if overrideParams != {}:
            log.critical("Override: %s" % overrideParams)
            self.override = True
            self.initialiseOverride()
        else:
            self.siteCfg = loadSiteLocalConfig()
            self.initialiseSiteConf()

    def stageFile(self, fileToStage, stageOut=True):
        """
        _stageFile_

        Use call to invoke transfers (either in or out)
        input:
            fileToStage: a dict containing at least:
                        LFN: the LFN for one end of the transfer, will be
                            mapped to a PFN before the transfer
                        PFN: (annoyingly) Not a PFN, but is the local filename
                            for the file when it's transferred to/from
            stageOut: boolean for if the file is staged in or out
        output:
            dict from fileToStage with PFN, PNN, StageOutCommand added

        I'm not entirely sure that StageOutCommand makes sense, but I don't want to break old code
        -AMM 6/30/2010

        """

        log.info("Working on file: %s" % fileToStage['LFN'])
        lfn = fileToStage['LFN']
        localFileName = fileToStage['PFN']
        self.firstException = None

        log.info("Beginning %s" % ('StageOut' if stageOut else 'StageIn'))

        # generate list of stageout methods we will try
        stageOutMethods = [self.defaultMethod]
        stageOutMethods.extend(self.fallbacks)

        # loop over all the different methods. This unifies regular and fallback stuff. Nice.
        methodCounter = 0
        for currentMethod in stageOutMethods:
            methodCounter += 1
            # the PFN that is received here is mapped from the LFN
            log.info("Getting transfer details for %s LFN %s" % (currentMethod, lfn))
            (pnn, command, options, pfn, protocol) = \
                self.getTransferDetails(lfn, currentMethod)
            log.info("Using PNN:     %s" % pnn)
            log.info("Command:       %s" % command)
            log.info("Options:       %s" % options)
            log.info("Protocol:      %s" % protocol)
            log.info("Mapped LFN:    %s" % lfn)
            log.info("    to PFN:    %s" % pfn)
            log.info("LocalFileName: %s" % localFileName)
            newPfn = self._doTransfer(currentMethod, methodCounter, localFileName, pfn, stageOut)
            if newPfn:
                log.info("Transfer succeeded: %s" % fileToStage)
                fileToStage['PFN'] = newPfn
                fileToStage['PNN'] = pnn
                fileToStage['StageOutCommand'] = command
                self.completedFiles[fileToStage['LFN']] = fileToStage
                return fileToStage
            else:
                # transfer method didn't work, go to next one
                continue
        # if we're here, then nothing worked. transferfail.
        log.error("Error in stageout")
        if self.firstException:
            raise self.firstException
        else:
            raise StageOutError("Error in stageout, this has been logged in the logs")

    def deleteLFN(self, lfn):
        """
        attempts to delete a file. will raise if none of the methods work, returns details otherwise
        """
        log.info("Beginning to delete %s" % 'lfn')
        retval = {}
        # generate list of stageout methods we will try
        stageOutMethods = [self.defaultMethod]
        stageOutMethods.extend(self.fallbacks)

        # loop over all the different methods. This unifies regular and fallback stuff. Nice.
        methodCounter = 0
        for currentMethod in stageOutMethods:
            methodCounter += 1
            (pnn, command, options, pfn, protocol) = \
                self.getTransferDetails(lfn, currentMethod)

            retval = {'LFN': lfn,
                      'PFN': pfn,
                      'PNN': pnn}

            log.info("Attempting deletion method %s" % (methodCounter,))
            log.info("Current method information: %s" % currentMethod)

            try:
                deleteSlave = retrieveStageOutImpl(command, useNewVersion=True)
            except RegistryError:
                deleteSlave = retrieveStageOutImpl(command, useNewVersion=False)
                logging.error("Tried to load stageout backend %s, a new version isn't there yet" % command)
                logging.error("Will try to fall back to the oldone, but it's really best to redo it")
                logging.error("Here goes...")
                deleteSlave.removeFile(pfn)
                return retval

            # do the delete. The implementation is responsible for its own verification
            try:
                deleteSlave.doDelete(pfn, pnn, command, options, protocol)
            except StageOutError as ex:
                log.info("Delete failed in an expected manner. Exception is:")
                log.info("%s" % str(ex))
                log.info(traceback.format_exc())
                if not self.firstException:
                    self.firstException = ex
                continue
            # note to people who think it's cheeky to catch exception after ranting against it:
            # this makes sense because no matter what the exception, we want to keep going
            # additionally, it prints out the proper backtrace so we can diagnose issues
            # AMM - 6/30/2010
            except Exception as ex:
                log.critical("Delete failed in an unexpected manner. Exception is:")
                log.critical("%s" % str(ex))
                log.info(traceback.format_exc())
                if not self.firstException:
                    self.firstException = ex
                continue

            # successful deletions make it here
            return retval

        # unseuccessful transfers make it here
        if self.firstException:
            raise self.firstException
        else:
            raise StageOutFailure("Could not delete", **retval)

    def initialiseSiteConf(self):
        """
        _initialiseSiteConf_

        Extract required information from site conf and TFC

        """
        implName = pnn = catalog = option = None
        try:
            implName = self.siteCfg.localStageOut.get("command")
            pnn = self.siteCfg.localStageOut.get("phedex-node")
            catalog = self.siteCfg.localStageOut.get("catalog")
            option = self.siteCfg.localStageOut.get('option', None)

        except Exception:
            log.critical('Either command, phedex-node or the catalog are missing from site-local-config.xml')
            log.critical('File operations cannot proceed like this')
            log.critical('command: %s phedex-node: %s catalog: %s' % (implName, pnn, catalog))
            raise
        try:
            self.tfc = self.siteCfg.trivialFileCatalog()
        except Exception:
            log.critical("TFC wasn't loaded, file operations cannot proceed")
            raise

        self.fallbacks = self.siteCfg.fallbackStageOut
        self.defaultMethod = {'command': implName,
                              'phedex-node': pnn,
                              'catalog': catalog}
        if option:
            self.defaultMethod['option'] = option

        log.info("Local Stage Out Implementation to be used is: %s" % implName)
        log.info("Local Stage Out PNN to be used is %s" % pnn)
        log.info("Local Stage Out Catalog to be used is %s" % catalog)
        log.info("Trivial File Catalog has been loaded:\n%s" % str(self.tfc))
        log.info("There are %s fallback stage out definitions" % len(self.fallbacks))
        for item in self.fallbacks:
            log.info("Fallback to : %s using: %s " % (item['phedex-node'], item['command']))

    def initialiseOverride(self):
        """
        _initialiseOverride_

        Extract required information from override.

        TODO: this should be merged with the initializeSiteConf function
        but I can't think of a nice way to do it

        """
        implName = pnn = lfn_prefix = None
        option = ""
        try:
            implName = self.overrideConf["command"]
            pnn = self.overrideConf["phedex-node"]
            lfn_prefix = self.overrideConf["lfn-prefix"]

        except Exception:
            log.critical('Either command, phedex-node, or the lfn-prefix are missing from the override')
            log.critical('File operations cannot proceed like this')
            log.critical('command: %s phedex-node: %s lfn-prefix: %s' % (implName, pnn, lfn_prefix))
            raise

        self.fallbacks = []
        self.defaultMethod = {'command': implName,
                              'phedex-node': pnn,
                              'lfn-prefix': lfn_prefix}
        if option:
            self.defaultMethod['option'] = option

        log.info("Note: We have been directed to use a StageOut override")
        log.info("Local Stage Out Implementation to be used is: %s" % implName)
        log.info("Local Stage Out PNN to be used is %s" % pnn)
        log.info("Local Stage Out lfn-prefix to be used is %s" % lfn_prefix)

    def getTransferDetails(self, lfn, currentMethod):
        """
        helper procedure to return the proper parameters to interact with the filesystem
        regardless of whether or not there's an override involved
        """

        if 'lfn-prefix' in currentMethod:
            pnn = currentMethod['phedex-node']
            command = currentMethod['command']
            options = currentMethod.get('option', None)
            pfn = "%s%s" % (currentMethod['lfn-prefix'], lfn)
            protocol = command
        else:
            pnn = self.siteCfg.localStageOut['phedex-node']
            command = self.siteCfg.localStageOut['command']
            options = self.siteCfg.localStageOut.get('option', None)
            pfn = self.searchTFC(lfn)
            protocol = self.tfc.preferredProtocol
        return pnn, command, options, pfn, protocol

    def stageIn(self, fileToStage):
        return self.stageFile(fileToStage, stageOut=False)

    def stageOut(self, fileToStage):
        return self.stageFile(fileToStage, stageOut=True)

    def _doTransfer(self, currentMethod, methodCounter, localFileName, pfn, stageOut):
        """
        performs a transfer using a selected method and retries.
        necessary because python doesn't have a good nested loop break syntax
        """

        (pnn, command, options, _, protocol) = \
            self.getTransferDetails(localFileName, currentMethod)

        # Swap directions if we're staging in
        if not stageOut:
            tempPfn = pfn
            pfn = localFileName
            localFileName = tempPfn

        for retryNumber in range(self.numberOfRetries + 1):
            log.info("Attempting transfer method %s, Retry number: %s" % (methodCounter, retryNumber))
            log.info("Current method information: %s" % currentMethod)

            try:
                stageOutSlave = retrieveStageOutImpl(command, useNewVersion=True, stagein=not stageOut)
            except RegistryError:
                stageOutSlave = retrieveStageOutImpl(command, useNewVersion=False, stagein=not stageOut)
                logging.error("Tried to load stageout backend %s, a new version isn't there yet" % command)
                logging.error("Will try to fall back to the oldone, but it's really best to redo it")
                logging.error("Here goes...")
                stageOutSlave(protocol, localFileName, pfn, options)
                return pfn

            # do the copy. The implementation is responsible for its own verification
            newPfn = None
            try:
                # FIXME add checksum stuff
                newPfn = stageOutSlave.doTransfer(localFileName, pfn, stageOut, pnn, command, options, protocol, None)
            except StageOutError as ex:
                log.info("Transfer failed in an expected manner. Exception is:")
                log.info("%s" % str(ex))
                log.info("Sleeping for %s seconds" % self.retryPauseTime)
                log.info(traceback.format_exc())
                time.sleep(self.retryPauseTime)
                if not self.firstException:
                    self.firstException = ex
                continue
            # note to people who think it's cheeky to catch exception after ranting against it:
            # this makes sense because no matter what the exception, we want to keep going
            # additionally, it prints out the proper backtrace so we can diagnose issues
            # AMM - 6/30/2010
            except Exception as ex:
                log.critical("Transfer failed in an unexpected manner. Exception is:")
                log.critical("%s" % str(ex))
                log.critical("Since this is an unexpected error, we are continuing to the next method")
                log.critical("and not retrying the same one")
                log.critical(traceback.format_exc())
                if not self.firstException:
                    self.firstException = ex
                break

            # successful transfers make it here
            return newPfn
        # unseuccessful transfers make it here
        return False

    def cleanSuccessfulStageOuts(self):
        """
        _cleanSucessfulStageOuts_

        In the event of a failed stage out, this method can be called to cleanup the
        files that may have previously been staged out so that the job ends in a clear state
        of failure, rather than a partial success


        """
        for lfn in self.completedFiles:
            log.info("Cleaning out file: %s\n" % lfn)
            try:
                self.deleteLFN(lfn)
            except StageOutFailure as ex:
                log.info("Failed to cleanup staged out file after error:")
                log.info(" %s\n%s" % (lfn, str(ex)))
                log.info(traceback.format_exc())

    def searchTFC(self, lfn):
        """
        _searchTFC_

        Search the Trivial File Catalog for the lfn provided,
        if a match is made, return the matched PFN

        """
        if self.tfc is None:
            log.info("Trivial File Catalog not available to match LFN:")
            log.info(lfn)
            return None
        if self.tfc.preferredProtocol is None:
            log.info("Trivial File Catalog does not have a preferred protocol")
            log.info("which prevents local stage out for:")
            log.info(lfn)
            return None

        pfn = self.tfc.matchLFN(self.tfc.preferredProtocol, lfn)
        if pfn is None:
            log.info("Unable to map LFN to PFN:")
            log.info("LFN: %s" % lfn)
            return None

        log.info("LFN to PFN match made:")
        log.info("LFN: %s\nPFN: %s\n" % (lfn, pfn))
        return pfn


# wrapper classes for compatibility
class StageInMgr(FileManager):
    def __init__(self, numberOfRetries=30, retryPauseTime=60, **overrideParams):
        FileManager.__init__(self, numberOfRetries=numberOfRetries, retryPauseTime=retryPauseTime, **overrideParams)

    def __call__(self, fileToStage):
        """
        stages in a file, fileToStage is a dict with at least the LFN key
        the dict will be modified and returned, or an exception will be raised
        """
        return self.stageIn(fileToStage)


class StageOutMgr(FileManager):
    def __init__(self, numberOfRetries=30, retryPauseTime=60, **overrideParams):
        FileManager.__init__(self, numberOfRetries=numberOfRetries, retryPauseTime=retryPauseTime, **overrideParams)

    def __call__(self, fileToStage):
        """
        stages out a file, fileToStage is a dict with at least the LFN key
        the dict will be modified and returned, or an exception will be raised
        """
        log.info("StageOutMgr called with file: %s" % fileToStage)
        return self.stageOut(fileToStage)


class DeleteMgr(FileManager):
    def __init__(self, numberOfRetries=30, retryPauseTime=60, **overrideParams):
        FileManager.__init__(self, numberOfRetries=numberOfRetries, retryPauseTime=retryPauseTime, **overrideParams)

    def __call__(self, fileToDelete):
        """
        stages out a file, fileToStage is a dict with at least the LFN key
        the dict will be modified and returned, or an exception will be raised
        """
        if 'LFN' not in fileToDelete:
            raise StageOutFailure('LFN not provided to deleteLFN')
        return self.deleteLFN(fileToDelete['LFN'])



#!/usr/bin/env python
"""
_AcquireFiles_

Oracle implementation of Subscription.GetAcquiredFiles
"""

from WMCore.WMBS.MySQL.Subscriptions.GetAcquiredFiles import GetAcquiredFiles \
     as GetAcquiredFilesMySQL

class GetAcquiredFiles(GetAcquiredFilesMySQL):
    pass



#!/usr/bin/env python
"""
_KillWorkflow_

MySQL implementation of Subscriptions.KillWorkflow
"""

from WMCore.Database.DBFormatter import DBFormatter

class KillWorkflow(DBFormatter):
    """
    _KillWorkflow_

    Mark all files that are not complete/failed and belong to a particular
    workflow as complete so they get clean up.  Ignore Cleanup and
    LogCollect subscriptions as we still want those to run.
    """
    sql = """INSERT INTO wmbs_sub_files_complete (subscription, fileid)
               SELECT wmbs_subscription.id, wmbs_fileset_files.fileid
                      FROM wmbs_workflow
                 INNER JOIN wmbs_subscription ON
                   wmbs_workflow.id = wmbs_subscription.workflow AND
                   wmbs_subscription.subtype IN
                    (SELECT id FROM wmbs_sub_types
                     WHERE name != 'Cleanup' AND name != 'LogCollect')
                 INNER JOIN wmbs_fileset_files ON
                   wmbs_subscription.fileset = wmbs_fileset_files.fileset
                 LEFT OUTER JOIN wmbs_sub_files_complete ON
                   wmbs_subscription.id = wmbs_sub_files_complete.subscription AND
                   wmbs_fileset_files.fileid = wmbs_sub_files_complete.fileid
                 LEFT OUTER JOIN wmbs_sub_files_failed ON
                   wmbs_subscription.id = wmbs_sub_files_failed.subscription AND
                   wmbs_fileset_files.fileid = wmbs_sub_files_failed.fileid
               WHERE wmbs_sub_files_complete.fileid IS Null AND
                     wmbs_sub_files_failed.fileid IS Null AND
                     wmbs_workflow.name = :workflowname"""

    delAcq = """DELETE FROM wmbs_sub_files_acquired WHERE subscription IN
                  (SELECT wmbs_subscription.id FROM wmbs_workflow
                     INNER JOIN wmbs_subscription ON
                       wmbs_workflow.id = wmbs_subscription.workflow AND
                       wmbs_workflow.name = :workflowname AND
                       wmbs_subscription.subtype IN
                         (SELECT id FROM wmbs_sub_types
                          WHERE name != 'Cleanup' AND name != 'LogCollect'))"""

    delAva = """DELETE FROM wmbs_sub_files_available WHERE subscription IN
                  (SELECT wmbs_subscription.id FROM wmbs_workflow
                     INNER JOIN wmbs_subscription ON
                       wmbs_workflow.id = wmbs_subscription.workflow AND
                       wmbs_workflow.name = :workflowname AND
                       wmbs_subscription.subtype IN
                         (SELECT id FROM wmbs_sub_types
                          WHERE name != 'Cleanup' AND name != 'LogCollect'))"""

    def execute(self, workflowName, conn = None, transaction = False):
        self.dbi.processData(self.sql, {"workflowname": workflowName},
                             conn = conn, transaction = transaction)
        self.dbi.processData(self.delAcq, {"workflowname": workflowName},
                             conn = conn, transaction = transaction)
        self.dbi.processData(self.delAva, {"workflowname": workflowName},
                             conn = conn, transaction = transaction)
        return



"""
_WorkUnit.Delete_

Oracle implementation of WorkUnit.Delete
"""

from __future__ import absolute_import, division, print_function

from WMCore.WMBS.MySQL.WorkUnit.Delete import Delete as DeleteMySQL


class Delete(DeleteMySQL):
    """
    _WorkUnit.Delete_

    Oracle implementation of WorkUnit.Delete
    """

    pass



#!/usr/bin/env python
"""
_ListDataset_

MySQL implementation of DBSBuffer.ListDataset
"""




from WMCore.Database.DBFormatter import DBFormatter

class ListDataset(DBFormatter):
    """
    _ListDataset_

    Retrieve information about a dataset in the DBSBuffer.  This is mostly used
    by the unit tests.
    """
    sql = """SELECT id, path, acquisition_era, processing_ver,
                    valid_status, global_tag, parent, prep_id FROM dbsbuffer_dataset
               WHERE path = :path"""

    sqlID = """SELECT id, path, acquisition_era, processing_ver,
                    valid_status, global_tag, parent, prep_id FROM dbsbuffer_dataset
                 WHERE id = :id"""

    def execute(self, datasetID = None, datasetPath = None, conn = None,
                transaction = False):
        """
        _execute_

        Either the dataset's ID or path must be specified.
        """
        if datasetID == None:
            binds = {"path": datasetPath}
            result = self.dbi.processData(self.sql, binds, conn = conn,
                                          transaction = transaction)
        else:
            binds = {"id": datasetID}
            result = self.dbi.processData(self.sqlID, binds, conn = conn,
                                          transaction = transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
_LoadFromSpecOwner_

MySQL implementation of Workflow.LoadFromSpecOwner
"""

from WMCore.Database.DBFormatter import DBFormatter

class LoadFromSpecOwner(DBFormatter):
    sql = """SELECT wmbs_workflow.id, wmbs_workflow.spec, wmbs_workflow.name,
                    wmbs_users.cert_dn as dn, wmbs_users.owner as owner,
                    wmbs_users.grp as grp, wmbs_users.group_name as vogrp,
                    wmbs_users.role_name as vorole, wmbs_workflow.task,
                    wmbs_workflow.type, wmbs_workflow.priority
             FROM wmbs_workflow
             INNER JOIN wmbs_users ON
               wmbs_workflow.owner = wmbs_users.id
             WHERE wmbs_workflow.spec = :spec AND wmbs_workflow.task = :task AND
               wmbs_users.cert_dn = :cert_dn"""

    def execute(self, spec, dn, task, conn = None,
                transaction = False):
        result = self.dbi.processData(self.sql, {"spec": spec, "cert_dn": dn,
                                                 "task": task},
                                      conn = conn, transaction = transaction)
        return self.formatDict(result)[0]



#!/usr/bin/env python
"""
_RunBased_

Run based splitting algorithm that will produce a set of jobs for files in a
(complete) run. Produces files one run at a time (run:JobGroup) - if the algo
created jobs over more than one run would have confusion in the JobGroup
creation and/or tracking.

If file spans a run will need to create a mask for that file.
"""




from builtins import range
from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.DataStructs.Fileset import Fileset
from WMCore.Services.UUIDLib import makeUUID

class RunBased(JobFactory):
    def algorithm(self, *args, **kwargs):
        """
        _algorithm_

        Implement run splitting algorithm. Assumes that if a file with a run is
        present, then all files for that run are also present.

        kwargs can take:
        files_per_job - e.g. 20 - Number of files per each split job
        """
        filesPerJob = int(kwargs.get("files_per_job", 300))
        requireRunClosed = bool(kwargs.get("require_run_closed", False))

        #baseName = makeUUID()

        # Select all primary files for the first present run
        curRun = None
        primaryFiles = []
        #The current objective of this code is to find all runs in
        #a fileset, and then for each run, create a jobGroup
        #in each jobGroup have a list of jobs containing all the
        #files for that run.
        #If files have more then one run, sort that file with
        #the lowest run
        #In future, mask these files?

        runDict = {}

        locationDict = self.sortByLocation()

        for location in locationDict:
            fileList = locationDict[location]
            for f in fileList:

                #If it is a WMBS object, load all data
                if hasattr(f, "loadData"):
                    f.loadData()

                #Die if there are no runs
                if len(f['runs']) < 1:
                    msg = "File %s claims to contain %s runs!" %(f['lfn'], len(f['runs']))
                    raise RuntimeError(msg)

                #First we need to pick the lowest run
                runList = []
                for r in f['runs']:
                    runList.append(r.run)

                run = min(runList)

                #If we don't have the run, we need to add it
                if run not in runDict:
                    runDict[run] = []

                runDict[run].append(f)


            for run in runDict:
                #Find the runs in the dictionary we assembled and split the files in them

                self.newGroup()
                baseName = makeUUID()

                #Now split them into sections according to files per job
                while len(runDict[run]) > 0:
                    jobFiles = []
                    for i in range(filesPerJob):
                        #Watch out if your last job has less then the full number of files
                        if len(runDict[run]) > 0:
                            jobFiles.append(runDict[run].pop())

                    # Create the job
                    currentJob = self.newJob('%s-%s' % (baseName, len(self.currentGroup.newjobs)),
                                             files = jobFiles)



#!/usr/bin/env python
"""
_SetLocationByLFN_

MySQL implementation of Files.SetLocationByLFN
"""

from WMCore.Database.DBFormatter import DBFormatter

from builtins import str, bytes

class SetLocationByLFN(DBFormatter):
    sql = """INSERT INTO wmbs_file_location (fileid, pnn)
                 SELECT wfd.id, wpnn.id
                 FROM wmbs_pnns wpnn, wmbs_file_details wfd
                 WHERE wpnn.pnn = :location
                 AND wfd.lfn = :lfn"""

    def getBinds(self, lfn=None, location=None):
        if isinstance(lfn, (str, bytes)):
            return {'lfn': lfn, 'location': location}
        elif isinstance(lfn, (list, set)):
            binds = []
            for bind in lfn:
                binds.append(bind)
            return binds

    def execute(self, lfn, location=None, conn=None, transaction=None):
        """
        Set location PNN by LFN

        """
        binds = self.getBinds(lfn, location)

        self.dbi.processData(self.sql, binds, conn=conn, transaction=transaction)
        return



#!/usr/bin/env python
"""
_GetState_

MySQL implementation of Jobs.GetState
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

from future.utils import listvalues

class GetState(DBFormatter):
    """
    _GetState_

    Given a job ID, get the state of a current job.
    """
    sql = "SELECT name FROM wmbs_job_state WHERE id = (SELECT state FROM wmbs_job wj WHERE wj.ID = :jobid)"

    def format(self, results):
        """
        _formatDict_

        Return a name from MySQL
        """

        if len(results) == 0:
            return False
        else:
            return listvalues(results[0].fetchall()[0])[0]


    def execute(self, id, conn = None, transaction = False):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        """
        result = self.dbi.processData(self.sql, {"jobid": id}, conn = conn,
                                      transaction = transaction)

        return self.format(result)



#!/usr/bin/env python
"""
_WorkUnit.Add_

MySQL implementation of WorkUnit.Add
"""

from __future__ import absolute_import, division, print_function

from WMCore.Database.DBFormatter import DBFormatter


class Add(DBFormatter):
    """
    _WorkUnit.Add_

    MySQL implementation of WorkUnit.Add
    """

    wuSQL = ('INSERT INTO wmbs_workunit (taskid, retry_count, last_unit_count, last_submit_time, status)'
             ' VALUES (:taskid, :retry_count, :last_unit_count, :last_submit_time, :status)')

    assocSQL = ('INSERT INTO wmbs_frl_workunit_assoc (workunit, firstevent, lastevent, fileid, run, lumi)'
                ' VALUES (LAST_INSERT_ID(), :firstevent, :lastevent, :fileid, :run, :lumi)')

    def execute(self, taskid=None, retry_count=0, last_unit_count=0, last_submit_time=0,
                status=0, first_event=0, last_event=0,
                fileid=None, run=None, lumi=None,
                conn=None, transaction=False):
        """
        Add the parts of the object that go into workunit and into the association table
        """

        # Add the part for the work unit table
        if taskid:
            binds = {'taskid': taskid, 'retry_count': retry_count, 'last_unit_count': last_unit_count,
                     'last_submit_time': last_submit_time, 'status': status}
            self.dbi.processData(self.wuSQL, binds, conn=conn, transaction=transaction)
        else:
            raise RuntimeError("No way to create work units without task ID or name")

        # And add the association information using the primary key from the first insertion
        binds = {'firstevent': first_event, 'lastevent': last_event, 'fileid': fileid, 'run': run, 'lumi': lumi}
        self.dbi.processData(self.assocSQL, binds, conn=conn, transaction=transaction)

        return



"""
    This module is used to fork the current process into a daemon.
    Almost none of this is necessary (or advisable) if your daemon
    is being started by inetd. In that case, stdin, stdout and stderr are
    all set up for you to refer to the network connection, and the fork()s
    and session manipulation should not be done (to avoid confusing inetd).
    Only the chdir() and umask() steps remain as useful.
    References:
        UNIX Programming FAQ
            1.7 How do I get my program to act like a daemon?
                http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        Advanced Programming in the Unix Environment
            W. Richard Stevens, 1992, Addison-Wesley, ISBN 0-201-56317-7.

    History:
      2001/07/10 by Jorgen Hermann
      2002/08/28 by Noah Spurrier
      2003/02/24 by Clark Evans

      http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66012
"""
from __future__ import print_function

from builtins import str
import logging
import os
import sys
import time
from logging.handlers import RotatingFileHandler
from xml.dom.minidom import Document

# File mode creation mask of the daemon.
UMASK = 0o022


def daemonize(stdout='/dev/null', stderr=None, stdin='/dev/null',
              workdir=None, startmsg='started with pid %s',
              keepParent=False):
    """
        This forks the current process into a daemon.
        The stdin, stdout, and stderr arguments are file names that
        will be opened and be used to replace the standard file descriptors
        in sys.stdin, sys.stdout, and sys.stderr.
        These arguments are optional and default to /dev/null.
        Note that stderr is opened unbuffered, so
        if it shares a file with stdout then interleaved output
        may not appear in the order that you expect.
    """
    # Do first fork.
    try:
        pid = os.fork()
        if pid > 0:
            if not keepParent:
                os._exit(0)  # Exit first parent.
            return pid
    except OSError as e:
        sys.stderr.write("fork #1 failed: (%d) %s\n" % (e.errno, e.strerror))
        print("fork #1 failed: (%d) %s\n" % (e.errno, e.strerror))
        sys.exit(1)
    # Decouple from parent environment.
    os.chdir("/")
    os.umask(UMASK)
    os.setsid()

    # Do second fork.
    try:
        pid = os.fork()
        if pid > 0:
            os._exit(0)  # Exit second parent.
    except OSError as e:
        sys.stderr.write("fork #2 failed: (%d) %s\n" % (e.errno, e.strerror))
        print("fork #2 failed: (%d) %s\n" % (e.errno, e.strerror))
        sys.exit(1)

    # Open file descriptors and print start message
    if not stderr:
        stderr = stdout
    si = open(stdin, 'r')
    so = open(stdout, 'a+')
    se = open(stderr, 'a+')
    pid = str(os.getpid())
    sys.stderr.write("\n%s\n" % startmsg % pid)
    sys.stderr.flush()
    if workdir:
        # open(pidfile,'w+').write("%s\n" % pid)
        # Since the current working directory may be a mounted filesystem, we
        # avoid the issue of not being able to unmount the filesystem at
        # shutdown time by changing it to the root directory.
        os.chdir(workdir)
        # We probably don't want the file mode creation mask inherited from
        # the parent, so we give the child complete control over permissions.
        os.umask(UMASK)

        xmlDoc = Document()
        daemon = xmlDoc.createElement("Daemon")
        processId = xmlDoc.createElement("ProcessID")
        processId.setAttribute("Value", str(os.getpid()))
        daemon.appendChild(processId)

        parentProcessId = xmlDoc.createElement("ParentProcessID")
        parentProcessId.setAttribute("Value", str(os.getppid()))
        daemon.appendChild(parentProcessId)

        processGroupId = xmlDoc.createElement("ProcessGroupID")
        processGroupId.setAttribute("Value", str(os.getpgrp()))
        daemon.appendChild(processGroupId)

        userId = xmlDoc.createElement("UserID")
        userId.setAttribute("Value", str(os.getuid()))
        daemon.appendChild(userId)

        effectiveUserId = xmlDoc.createElement("EffectiveUserID")
        effectiveUserId.setAttribute("Value", str(os.geteuid()))
        daemon.appendChild(effectiveUserId)

        groupId = xmlDoc.createElement("GroupID")
        groupId.setAttribute("Value", str(os.getgid()))
        daemon.appendChild(groupId)

        effectiveGroupId = xmlDoc.createElement("EffectiveGroupID")
        effectiveGroupId.setAttribute("Value", str(os.getegid()))
        daemon.appendChild(effectiveGroupId)

        with open("Daemon.xml", "w") as props:
            props.write(daemon.toprettyxml())

    # Redirect standard file descriptors.
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())
    return 0


def test():
    """
        This is an example main function run by the daemon.
        This prints a count and timestamp once per second.
    """
    logFile = os.path.join("/tmp/daemon.logging")
    logHandler = RotatingFileHandler(logFile, "a", 1000000000, 3)
    logFormatter = logging.Formatter("%(asctime)s:%(levelname)s:%(module)s:%(message)s")
    logHandler.setFormatter(logFormatter)
    logging.getLogger().addHandler(logHandler)
    logging.getLogger().setLevel(logging.INFO)
    logging.info("Message to log file")

    sys.stdout.write('Message to stdout...')
    sys.stderr.write('Message to stderr...')
    c = 0
    while True:
        sys.stdout.write('%d: %s\n' % (c, time.ctime(time.time())))
        logging.info('%d: %s\n', c, time.ctime(time.time()))
        sys.stdout.flush()
        c = c + 1
        time.sleep(1)
    logging.info('>>>Starting:<<<')


def createDaemon(workdir, keepParent=False):
    """
    This is a wrapper over the new daemon methods.
    That follows the same interface.

    """
    pidfile = os.path.join(workdir, 'Daemon.xml')
    startmsg = 'started with pid %s'
    try:
        pf = open(pidfile, 'r')
        pid = (pf.read().strip())
        pf.close()
    except IOError:
        pid = None
    if pid:
        mess = """
Start aborted since pid file '%s' exists.
Please kill process and remove file first.
If process is still running this file contains
information on that.
"""
        print(mess % pidfile)
        sys.exit(1)

    return daemonize(workdir=workdir, startmsg=startmsg, keepParent=keepParent)


if __name__ == "__main__":
    parent_id = createDaemon('/tmp', keepParent=False)
    if parent_id == 0:
        test()
    print('Kept parent: ' + str(parent_id))



#!/usr/bin/env python
"""
_KillWorkflow_

Oracle implementation of Jobs.KillWorkflow
"""

from WMCore.WMBS.MySQL.Jobs.KillWorkflow import KillWorkflow as KillWorkflowMySQL

class KillWorkflow(KillWorkflowMySQL):
    pass



#!/usr/bin/env python
"""
_Subscription.New_

MySQL implementation of Subscription.New
"""

import time

from WMCore.Database.DBFormatter import DBFormatter

class New(DBFormatter):
    """
    _New_

    Create a new subscription.  Add any files that exist in the fileset to the
    wmbs_sub_files_available table.
    """
    typesSQL = """INSERT IGNORE INTO wmbs_sub_types (name)
                    VALUES (:subtype)"""

    sql = """INSERT INTO wmbs_subscription (fileset, workflow, subtype,
                                            split_algo, last_update)
               SELECT :fileset, :workflow, id, :split_algo, :timestamp
                      FROM wmbs_sub_types WHERE name = :subtype"""

    sqlAvail = """INSERT INTO wmbs_sub_files_available (subscription, fileid)
                    SELECT wmbs_subscription.id, wmbs_fileset_files.fileid
                           FROM wmbs_fileset_files
                      INNER JOIN wmbs_subscription ON
                        wmbs_subscription.workflow = :workflow AND
                        wmbs_subscription.fileset = :fileset AND
                        wmbs_fileset_files.fileset = wmbs_subscription.fileset"""

    def execute(self, fileset = None, workflow = None, split_algo = "File",
                type = "Processing", conn = None, transaction = False):
        binds = {"fileset": fileset, "workflow": workflow, "subtype": type,
                 "split_algo": split_algo, "timestamp": int(time.time())}
        availBinds = {"fileset": fileset, "workflow": workflow}

        self.dbi.processData(self.typesSQL, {'subtype': type}, conn = conn,
                             transaction = transaction)
        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        self.dbi.processData(self.sqlAvail, availBinds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
MySQL implementation of File.SetParentage

Make the parentage link between two file lfns in bulk
"""

from WMCore.Database.DBFormatter import DBFormatter

class SetParentage(DBFormatter):
    sql = """INSERT IGNORE INTO wmbs_file_parent (child, parent)
             SELECT DISTINCT wfd1.id, wfd2.id
             FROM wmbs_file_details wfd1 INNER JOIN wmbs_file_details wfd2
             WHERE wfd1.lfn = :child
             AND wfd2.lfn = :parent
    """

    def execute(self, binds, conn = None, transaction = False):
        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_SubscriptionStatus_

Retrieve status information about subscriptions in WMBS.  This will return a
list of disctionaries with the following keys:
    percent_success
    percent_complete
    fileset_name
    workflow_name
    subscription_id

"""
from __future__ import division

from WMCore.Database.DBFormatter import DBFormatter

class SubscriptionStatus(DBFormatter):
    sql = """SELECT wmbs_subscription.id AS subscription_id,
                    wmbs_workflow.name AS workflow_name,
                    wmbs_fileset.name AS fileset_name,
                    job_info.job_state AS job_state,
                    job_info.job_count AS job_count,
                    job_info.num_success AS success_count FROM wmbs_subscription
               INNER JOIN wmbs_workflow ON
                 wmbs_subscription.workflow = wmbs_workflow.id
               INNER JOIN wmbs_fileset ON
                 wmbs_subscription.fileset = wmbs_fileset.id
               LEFT OUTER JOIN
                 (SELECT wmbs_jobgroup.subscription AS subscription,
                         COUNT(*) AS job_count,
                         SUM(wmbs_job.outcome) AS num_success,
                         wmbs_job_state.name AS job_state FROM wmbs_jobgroup
                    INNER JOIN wmbs_job ON
                      wmbs_jobgroup.id = wmbs_job.jobgroup
                    INNER JOIN wmbs_job_state ON
                      wmbs_job.state = wmbs_job_state.id
                  GROUP BY wmbs_jobgroup.subscription, wmbs_job_state.name) job_info ON
                 wmbs_subscription.id = job_info.subscription"""

    typeSql = """SELECT wmbs_subscription.id AS subscription_id,
                        wmbs_workflow.name AS workflow_name,
                        wmbs_fileset.name AS fileset_name,
                        job_info.job_state AS job_state,
                        job_info.job_count AS job_count,
                        job_info.num_success AS success_count FROM wmbs_subscription
                   INNER JOIN wmbs_workflow ON
                     wmbs_subscription.workflow = wmbs_workflow.id
                   INNER JOIN wmbs_fileset ON
                     wmbs_subscription.fileset = wmbs_fileset.id
                   INNER JOIN wmbs_sub_types ON
                     wmbs_subscription.subtype = wmbs_sub_types.id
                   LEFT OUTER JOIN
                     (SELECT wmbs_jobgroup.subscription AS subscription,
                             COUNT(*) AS job_count,
                             SUM(wmbs_job.outcome) AS num_success,
                             wmbs_job_state.name AS job_state FROM wmbs_jobgroup
                        INNER JOIN wmbs_job ON
                          wmbs_jobgroup.id = wmbs_job.jobgroup
                        INNER JOIN wmbs_job_state ON
                          wmbs_job.state = wmbs_job_state.id
                      GROUP BY wmbs_jobgroup.subscription, wmbs_job_state.name) job_info ON
                     wmbs_subscription.id = job_info.subscription
                 WHERE wmbs_sub_types.name = :subscriptionType"""

    def format(self, result):
        """
        _format_

        Format the results of the query into something reasonable.  The query
        will return a row for each subscription/job type combination so we need
        to condense that down and return a single dictionary with information on
        job status for each subscription.
        """
        results = DBFormatter.formatDict(self, result)

        workflows = {}
        for result in results:
            if result["workflow_name"] not in workflows:
                workflows[result["workflow_name"]] = {}

            workflowDict = workflows[result["workflow_name"]]
            if result["fileset_name"] not in workflowDict:
                workflowDict[result["fileset_name"]] = {"success": 0, "running": 0,
                                                       "failure": 0,
                                                       "subId": result["subscription_id"]}

            filesetDict = workflowDict[result["fileset_name"]]
            if result["job_state"] in ("exhausted", "cleanout", "success", "jobfailed"):
                if result["success_count"] != None:
                    filesetDict["success"] += result["success_count"]
                    filesetDict["failure"] += result["job_count"] - result["success_count"]
            else:
                if result["success_count"] != None:
                    filesetDict["running"] += result["job_count"]

        results = []
        for workflowName in workflows:
            for filesetName in workflows[workflowName]:
                success = workflows[workflowName][filesetName]["success"]
                failure = workflows[workflowName][filesetName]["failure"]
                running = workflows[workflowName][filesetName]["running"]

                if success + failure + running == 0:
                    percentComplete = 0
                    percentSuccess = 0
                elif success + failure == 0:
                    percentComplete = int((success + failure) / (success + failure + running) * 100)
                    percentSuccess = 0
                else:
                    percentComplete = int((success + failure) / (success + failure + running) * 100)
                    percentSuccess = int(success / (success + failure + running) * 100)

                subId = workflows[workflowName][filesetName]["subId"]

                results.append({"subscription_id": subId,
                                "workflow_name": workflowName,
                                "fileset_name": filesetName,
                                "percent_complete": percentComplete,
                                "percent_success": percentSuccess})
        return results

    def execute(self, subscriptionType = None, conn = None,
                transaction = False):
        if subscriptionType == None:
            result = self.dbi.processData(self.sql, conn = conn,
                                          transaction = transaction)
        else:
            bindVars = {"subscriptionType": subscriptionType}
            result = self.dbi.processData(self.typeSql, bindVars, conn = conn,
                                          transaction = transaction)

        return self.format(result)



#!/usr/bin/env python
"""
_ListWorkloadsForTaskSite_

Oracle implementation of ResourceControl.ListWorkloadsForTaskSite
"""




from WMCore.ResourceControl.MySQL.ListWorkloadsForTaskSite \
     import ListWorkloadsForTaskSite as MySQLListWorkloadsForTaskSite

class ListWorkloadsForTaskSite(MySQLListWorkloadsForTaskSite):
    pass



#!/bin/env python


"""
_LinearAlgo_

Straight lines to the end
"""
import os.path
import logging

from WMCore.FwkJobReport.Report                     import Report
from WMComponent.RetryManager.PlugIns.RetryAlgoBase import RetryAlgoBase

class ProcessingAlgo(RetryAlgoBase):
    """
    _ProcessingAlgo_

    Test jobs for two specific error conditions:

    If job has an exit code within a list of exit codes, only retry once
      set by ProcessingAlgoOneMoreErrorCodes
    If a job ran for over 24 hours, only retry once
      set by ProcessingAlgoMaxRuntime

    This is done by setting the retry_count to be what we think is the maxRetry count - 1
    """

    def __init__(self, config):

        # Init basics
        RetryAlgoBase.__init__(self, config)

        self.maxRetries = 3

        self.maxRunTime = self.getAlgoParam('default', 'MaxRunTime', 24 * 3600)
        self.exitCodes  = self.getAlgoParam('default', 'OneMoreErrorCodes', [])

        # Try to get the actual number of max retries, but don't mind
        # if this is a test implementation without the full config
        try:
            self.maxRetries = self.config.ErrorHandler.maxRetries
        except:
            logging.debug("No ErrorHandler component passed to RetryManager.ProcessingAlgo - hope this is a test.")
            pass

        return

    def isReady(self, job, cooloffType):
        """
        Actual function that does the work

        """

        if cooloffType == 'create' or cooloffType == 'submit':
            # Can't really do anything with these: resubmit
            return True

        # Run this to get the errors in the actual job
        try:
            report     = Report()
            reportPath = os.path.join(job['cache_dir'], "Report.%i.pkl" % job['retry_count'])
            report.load(reportPath)
        except:
            # If we're here, then the FWJR doesn't exist.
            # Give up, run it again
            return True

        # Set oneMore flag to be False
        oneMore = False

        # Find startTime, stopTime
        times = report.getFirstStartLastStop()
        startTime = times['startTime']
        stopTime  = times['stopTime']

        if startTime == None or stopTime == None:
            # Well, then we have a problem.
            # There is something very wrong with this job, nevertheless we don't know what it is.
            # Rerun, and hope the times get written the next time around.
            logging.error("No start, stop times for steps")
            return True

        if stopTime - startTime > self.maxRunTime:
            logging.error("Job only allowed to run one more time due to ProcessingAlgo.maxRunTime")
            oneMore = True

        if report.getExitCode() in self.exitCodes:
            logging.error("Job only allowed to run one more time due to ProcessingAlgo.exitCodes")
            oneMore = True


        # Reset the retry time
        if oneMore:
            job['retry_count'] = max(self.maxRetries - 1, job['retry_count'])
            job.save()
            # Hope this gets passed back by reference

        return True



"""
_GetHeartbeatInfo_

Oracle implementation of GetHeartbeatInfo
"""

__all__ = []



from WMCore.Agent.Database.MySQL.GetHeartbeatInfo import GetHeartbeatInfo \
     as GetHeartbeatInfoMySQL

class GetHeartbeatInfo(GetHeartbeatInfoMySQL):
    pass



#!/usr/bin/env python

"""
_Watchdog_

This cleverly named object is the thread that handles the monitoring of individual jobs
"""

import logging
import os
import os.path
import threading
import traceback

from PSetTweaks.WMTweak import resizeResources
from WMCore.WMException import WMException
from WMCore.WMFactory import WMFactory


class WatchdogException(WMException):
    """
    _StepFactortyException_

    It's like an exception class that does nothing

    """
    pass


class Watchdog(threading.Thread):
    """
    Watchdog

    It's like a watchdog!  It watches things, and then it barks at them, curls up, and goes to sleep
    Or possibly monitors them and records what they do.  It's a bit of a crapshoot at this point.
    """

    def __init__(self, logPath=None, config=None):
        threading.Thread.__init__(self)
        self.doMonitoring = True
        self._Finished = threading.Event()
        self._EndOfJob = threading.Event()
        self._NewTask = threading.Event()
        self._JobKilled = threading.Event()
        self._RunUpdate = threading.Event()
        self._Interval = 120.0
        self._Monitors = []

        # Right now we join this, because we don't know
        # Where we'll be when we need this.
        self.logPath = os.path.join(os.getcwd(), logPath)

        self.factory = WMFactory(self.__class__.__name__,
                                 "WMCore.WMRuntime.Monitors")

    def setupMonitors(self, task, wmbsJob):
        logging.info("In Watchdog.setupMonitors")
        if not hasattr(task.data, 'watchdog'):
            msg = "Could not find watchdog in spec"
            logging.error(msg)
            # I don't think this is necessarily fatal
            return
        if not hasattr(task.data.watchdog, 'monitors'):
            msg = "Watchdog has no monitors"
            logging.error(msg)
            # Probably not fatal either
            return
        if hasattr(task.data.watchdog, 'interval'):
            # Set the interval off the config
            self.setInterval(task.data.watchdog.interval)
        for monitor in task.data.watchdog.monitors:
            msg = "Initializing monitor %s" % monitor
            logging.info(msg)
            mon = self.loadMonitor(monitor)
            args = {}
            if hasattr(task.data.watchdog, monitor):
                # This should be a config section
                monitorArgs = getattr(task.data.watchdog, monitor)
                args = monitorArgs.dictionary_()
            if monitor == 'PerformanceMonitor' and args:
                # Apply tweaks to PerformanceMonitor only.
                # Scale resources according to the HTCondor runtime environment.
                origCores = 1
                for stepName in task.listAllStepNames():
                    sh = task.getStepHelper(stepName)
                    origCores = max(origCores, sh.getNumberOfCores())
                resources = {'cores': origCores}
                origMaxPSS = args.get('maxPSS', args.get('maxRSS'))
                if origMaxPSS:
                    resources['memory'] = origMaxPSS
                # Actually parses the HTCondor runtime
                resizeResources(resources)
                # We decided to only touch Watchdog settings if the number of cores changed.
                # (even if this means the watchdog memory is wrong for a slot this size).
                changedCores = origCores != resources['cores']
                # If we did base maxPSS off the memory in the HTCondor slot, subtract a bit
                # off the top so watchdog triggers before HTCondor does.
                # Add the new number of cores to the args such that PerformanceMonitor can see it
                args['cores'] = resources['cores']
                if changedCores:
                    if origMaxPSS:
                        args['maxPSS'] = resources['memory'] - 50

                logging.info("Watchdog modified: %s. Final settings:", changedCores)
                for k, v in args.items():
                    logging.info("  %s: %r", k, v)
            # Actually initialize the monitor variables
            mon.initMonitor(task=task, job=wmbsJob,
                            logPath=self.logPath, args=args)
            self._Monitors.append(mon)

        return

    def loadMonitor(self, monitorName):
        try:
            return self.factory.loadObject(monitorName)
        except WMException:
            msg = "WatchdogFactory Unable to load Object: %s" % monitorName
            logging.error(msg)
            raise WatchdogException(msg)
        except Exception as ex:
            msg = "Error creating object %s in WatchdogFactory:\n" % monitorName
            msg += str(ex)
            logging.error(msg)
            raise WatchdogException(msg)

    def setInterval(self, interval):
        """
        Set the monitor interval
        """
        logging.info("Set Watchdog interval to %s", interval)
        self._Interval = interval

    def disableMonitoring(self):
        """
        _disableMonitoring_

        Turn off active monitoring (periodicUpdate calls are disabled,
        event driven monitoring still occurs
        """
        self.doMonitoring = False

    def shutdown(self):
        """
        Shutdown the monitor.
        """
        logging.info("MonitorState: Shutdown called")
        self._Finished.set()
        return

    #  //=========notify Methods called by the ExecutionManager====
    # //
    # //  Start notification from the exe thread, this starts the
    #  //periodic updates of the monitor thread
    # //
    # //
    def notifyJobStart(self, task):
        """
        Start the job.
        """
        logging.info("MonitorThread: JobStarted")
        if self.doMonitoring:
            self.setDaemon(1)
            self.start()
        for monitor in self._Monitors:
            try:
                monitor.jobStart(task)
            except Exception as ex:
                msg = "Error in notifyJobStart for monitor class %s in Watchdog:\n" % monitor.__class__
                msg += str(ex)
                msg += str(traceback.format_exc())
                logging.error(msg)
                raise WatchdogException(msg)
        return

    #  //
    # // notify Monitors of new task start up
    # //
    def notifyStepStart(self, step):
        """
        notify Monitors of new task start up.
        """
        self._RunUpdate.set()
        for monitor in self._Monitors:
            try:
                monitor.stepStart(step)
            except Exception as ex:
                msg = "Error in notifyTaskStart for monitor class %s in Watchdog:\n" % monitor.__class__
                msg += str(ex)
                msg += str(traceback.format_exc())
                logging.error(msg)
                raise WatchdogException(msg)
        return

    #  //
    # // notify Monitors of task completion
    # //
    def notifyStepEnd(self, step, exitCode=0, stepReport=None):
        """
        notify Monitors of task completion.
        """
        self._RunUpdate.clear()
        for monitor in self._Monitors:
            try:
                monitor.stepEnd(step=step, stepReport=stepReport)
            except Exception as ex:
                msg = "Error in notifyTaskEnd for monitor class %s in Watchdog:\n" % monitor.__class__
                msg += str(ex)
                msg += str(traceback.format_exc())
                logging.error(msg)
                raise WatchdogException(msg)
        # print "Task Ended: %s with Exit Code:%s" % (task, exitCode)
        # self._MonMgr.taskEnd(task, exitCode)
        return

    #  //
    # // notify monitors of Job Completion, stops the periodic
    # //  updating
    def notifyJobEnd(self, task):
        """
        notify monitors of Job Completion, stops the periodic
        updating.
        """
        logging.info("MonitorThread: JobEnded")
        for monitor in self._Monitors:
            try:
                monitor.jobEnd(task)
            except Exception as ex:
                msg = "Error in notifyJobEnd for monitor class %s in Watchdog:\n" % monitor.__class__
                msg += str(ex)
                msg += str(traceback.format_exc())
                raise WatchdogException(msg)
        # self._MonMgr.jobEnd()
        self.shutdown()
        return

    #  //
    # //  Interrupt Notifiers
    # //   Job has been killed
    def notifyKillJob(self):
        """
        Interrupt Notifiers, Job has been killed.
        """
        logging.info("MonitorThread: JobKilled")
        for monitor in self._Monitors:
            try:
                monitor.jobKilled()
            except Exception as ex:
                msg = "Error in notifyKillJob for monitor class %s in Watchdog:\n" % monitor.__class__
                msg += str(ex)
                msg += str(traceback.format_exc())
                logging.error(msg)
                raise WatchdogException(msg)
        # self._MonMgr.jobKilled()
        self.shutdown()

    #  //
    # //  Task has been killed
    # //
    def notifyKillStep(self, step=None):
        """
        Task has been killed.
        """
        logging.info("MonitorThread: TaskKilled")
        for monitor in self._Monitors:
            try:
                monitor.stepKilled(step)
            except Exception as ex:
                msg = "Error in notifyKillTask for monitor class %s in Watchdog:\n" % monitor.__class__
                msg += str(ex)
                msg += str(traceback.format_exc())
                logging.error(msg)
                raise WatchdogException(msg)
                # self._MonMgr.taskKilled()

    # //
    # // Override Thread.run() to do the periodic update
    # //  of the MonitorState object and dispatch it to the monitors
    def run(self):
        """
        Override Thread.run() to do the periodic update
        of the MonitorState object and dispatch it to the monitors
        """
        while True:
            #  //
            # // shutdown signal
            # //
            if self._Finished.isSet():
                return

            # //
            # // Update State information only during a running task
            # //
            if self._RunUpdate.isSet():
                for monitor in self._Monitors:
                    try:
                        monitor.periodicUpdate()
                    except Exception as ex:
                        msg = "Error in periodicUpdate for monitor class %s in Watchdog:\n" % monitor.__class__
                        msg += str(ex)
                        msg += str(traceback.format_exc())
                        msg += "This is a CRITICAL error because this kills the monitoring.\n"
                        msg += "Terminate thread and retry.\n"
                        logging.error(msg)
                        # raise WatchdogException(msg)
                        # This one needs to be killed by itself
                        # since it's run by thread
                        os.abort()
                        # self._MonMgr.periodicUpdate()

            # time.sleep(self._Interval)
            self._Finished.wait(self._Interval)

    # //
    # // Load Monitor Objects based on Cfg settings passed
    # //  from Executor
    def initMonitorFwk(self, monitorCfg, updatorCfg):
        """
        _initMonitorFwk_

        Initialise the MonitorMgr object when this method is
        called from the Execution thread
        Load Monitor Objects based on Cfg settings passed from Executor.
        """
        # self._MonMgr.monitorConfig = monitorCfg
        # self._MonMgr.updatorConfig = updatorCfg
        # self._MonMgr.loadMonitors()
        # self._MonMgr.loadUpdators()
        return



#!/usr/bin/env python
"""
_MergeBySize_

"Splitting" algorithm to combine a fileset into merge job definitions
based on the size of the files

"""




from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.Services.UUIDLib import makeUUID
from WMCore.DataStructs.Fileset import Fileset



class MergeBySize(JobFactory):
    """
    _MergeBySize_

    Process a fileset to generate merge jobs based on size.
    Files will be combined based on size,  grouped by
    run,and sorted by lumisection

    an option is provided to include all files in overflow jobs


    """
    def algorithm(self, *args, **kwargs):
        """
        _algorithm_

        Implement merge algorithm for the subscription provided

        """
        fileset = list(self.subscription.availableFiles())

        mergeSize = int(kwargs['merge_size'])
        overflow  = bool(kwargs.get('all_files', False))
        fileset.sort()

        accumSize = 0
        jobFiles = Fileset()
        locationDict = self.sortByLocation()
        for location in locationDict:
            baseName = makeUUID()
            self.newGroup()
            for f in locationDict[location]:
                accumSize += f['size']
                jobFiles.addFile(f)
                if accumSize >= mergeSize:
                    self.newJob(name = '%s-%s' % (baseName, len(self.currentGroup.jobs) + 1),
                                      files = jobFiles)
                    self.currentJob["mask"].setMaxAndSkipEvents(-1, 0)
                    accumSize = 0
                    jobFiles = Fileset()

            if len(jobFiles) > 0:
                if overflow:
                    self.newJob(name = '%s-%s' % (baseName, len(self.currentGroup.jobs) + 1),
                                      files = jobFiles)
                    self.currentJob["mask"].setMaxAndSkipEvents(-1, 0)



#!/usr/bin/env python
"""
_NewWorkflow_

Oracle implementation of NewWorkflow
"""

from WMCore.WMBS.MySQL.Workflow.New import New as NewWorkflowMySQL

class New(NewWorkflowMySQL):
    sql = """insert into wmbs_workflow (id, spec, owner, name, task, type, alt_fs_close, priority)
             values (wmbs_workflow_SEQ.nextval, :spec, :owner, :name, :task, :type, :alt_fs_close, :priority)"""



#!/usr/bin/env python
"""
_GetOutputParentLFNs_

MySQL implementation of Jobs.GetOutputParentLFNs
"""




from WMCore.Database.DBFormatter import DBFormatter

class GetOutputParentLFNs(DBFormatter):
    inputSQL = """SELECT DISTINCT wmbs_file_details.lfn, wmbs_file_details.merged
                  FROM wmbs_file_details
                    INNER JOIN wmbs_job_assoc ON
                      wmbs_file_details.id = wmbs_job_assoc.fileid
                  WHERE wmbs_job_assoc.job = :job"""

    parentSQL = """SELECT DISTINCT wmbs_file_details.lfn FROM wmbs_file_details
                     INNER JOIN wmbs_file_parent ON
                       wmbs_file_details.id = wmbs_file_parent.parent
                     INNER JOIN wmbs_job_assoc ON
                       wmbs_file_parent.child = wmbs_job_assoc.fileid
                   WHERE wmbs_job_assoc.job = :job"""

    def format(self, results):
        """
        _format_

        Format the result of the query so that it is just a single list of LFNs.
        """
        results = DBFormatter.format(self, results)

        lfns = []
        for result in results:
            lfns.append(result[0])

        return lfns

    def execute(self, jobID, conn = None, transaction = False):
        result = self.dbi.processData(self.inputSQL, {"job": jobID}, conn = conn,
                                      transaction = transaction)

        formattedResult = DBFormatter.format(self, result)

        if len(formattedResult) == 0:
            return []

        if int(formattedResult[0][1]) == 0:
            # The input to the job consisted of unmerged files, so we'll need
            # to query for the parents of the job's input.
            result = self.dbi.processData(self.parentSQL, {"job": jobID}, conn = conn,
                                          transaction = transaction)

        return self.format(result)



#!/usr/bin/env python
"""
_JobStatusForMonitoring_

Oracle implementation for loading a job by scheduler status
"""

from WMCore.BossAir.MySQL.JobStatusForMonitoring \
     import JobStatusForMonitoring as MySQLJobStatusForMonitoring

class JobStatusForMonitoring(MySQLJobStatusForMonitoring):
    sql = """SELECT wwf.name as workflow, count(rj.wmbs_id) AS num_jobs,
                    st.name AS status, wl.plugin AS plugin, wu.cert_dn AS owner
               FROM bl_runjob rj
               INNER JOIN wmbs_users wu ON wu.id = rj.user_id
               INNER JOIN bl_status st ON rj.sched_status = st.id
               INNER JOIN wmbs_job wj ON wj.id = rj.wmbs_id
               INNER JOIN wmbs_jobgroup wjg ON wjg.id = wj.jobgroup
               INNER JOIN wmbs_subscription ws ON ws.id = wjg.subscription
               INNER JOIN wmbs_workflow wwf ON wwf.id = ws.workflow
               LEFT OUTER JOIN wmbs_location wl ON wl.id = wj.location
               WHERE rj.status = :complete
               GROUP BY wwf.name, wl.plugin, st.name, wu.cert_dn
    """



#!/usr/bin/env python
"""
_CheckInjectedWorkflow_

MySQL implementation of Workflow.CheckInjectedWorkflow

"""
__all__ = []

from WMCore.Database.DBFormatter import DBFormatter

class CheckInjectedWorkflow(DBFormatter):
    """
    Checks workflows to see if they have been fully injected
    """

    sql = """SELECT injected FROM wmbs_workflow WHERE name = :name"""

    def execute(self, name, conn = None, transaction = False):
        """
        Update the workflows to match their injected status

        """
        result = self.dbi.processData(self.sql, {"name": name}, conn = conn,
                                      transaction = transaction)

        d = self.formatDict(result)[0]

        if d.get('injected', 0) == 1:
            return True
        else:
            return False



#!/usr/bin/env python
"""
_ListSitesSlotsState_

Query the Resource Control database to get all the current sites,
running slots, pending slots and state.
"""
from WMCore.Database.DBFormatter import DBFormatter

class ListSitesSlotsState(DBFormatter):
    sql = """SELECT site_name, running_slots, pending_slots,
                 wlst.name AS state FROM wmbs_location
                 INNER JOIN wmbs_location_state wlst ON wlst.id = wmbs_location.state
           """

    def execute(self, conn = None, transaction = False):
        results = self.dbi.processData(self.sql, conn = conn, transaction = transaction)
        results = self.formatDict(results)

        dictResults = {}
        for elem in results:
            dictResults[elem['site_name']] = {'running_slots': elem['running_slots'],
                                              'pending_slots': elem['pending_slots'],
                                              'state': elem['state']}
        return dictResults



#!/usr/bin/env python
"""
_GetAvailableFilesNoLocations_

Oracle implementation of Subscription.GetAvailableFilesNoLocations
"""

from WMCore.WMBS.MySQL.Subscriptions.GetAvailableFilesNoLocations \
     import GetAvailableFilesNoLocations as GetAvailableFilesNoLocationsMySQL

class GetAvailableFilesNoLocations(GetAvailableFilesNoLocationsMySQL):
    pass



#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
# pylint: disable=E1101,C0103,R0902,E0602
# E0602: sneakily importing a lot of stuff, turn off for false positives

"""
Couch DB command line admin tool
"""
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import urllib.parse

__license__ = "GPL"
__maintainer__ = "Valentin Kuznetsov"
__email__ = "vkuznet@gmail.com"
__status__ = "Alpha"

# system modules
import os
import sys
import inspect
import traceback

# ipython modules
from   IPython import Release
import IPython.ipapi
import __main__


class PrintManager:
    def __init__(self):
        from IPython import ColorANSI
        self.term = ColorANSI.TermColors

    def print_red(self, msg):
        """print message using red color"""
        print(self.msg_red(msg))

    def print_green(self, msg):
        """print message using blue color"""
        print(self.msg_green(msg))

    def print_blue(self, msg):
        """print message using blue color"""
        print(self.msg_blue(msg))

    def msg_red(self, msg):
        """yield message using red color"""
        if not msg:
            msg = ''
        return self.term.Red + msg + self.term.Black

    def msg_green(self, msg):
        """yield message using green color"""
        if not msg:
            msg = ''
        return self.term.Green + msg + self.term.Black

    def msg_blue(self, msg):
        """yield message using blue color"""
        if not msg:
            msg = ''
        return self.term.Blue + msg + self.term.Black


def load_config(func_list=[]):
    """
    Defines default configuration for Couch DB. We need the following:
    URI - the Couch DB URI
    DB  - the Couch DB name
    DESIGN - the Couch DB design name
    DEBUG  - debug level, used to setup HTTPConnection debug level
    """
    msg = """
import os, re, sys, time, types, traceback, inspect
import urllib, urllib2, httplib
import json
from json import JSONDecoder, JSONEncoder
try:
    from path import path
except ImportError:
    pass
try:
    from ipipe import *
except ImportError:
    pass

# global variables
URI="http://localhost:5984"
DB="das"
DESIGN="dasadmin"
DEBUG=0
"""
    msg += "\n%s\n" % inspect.getsource(PrintManager)
    msg += "PM = PrintManager()"
    for func in func_list:
        msg += "\n%s\n" % inspect.getsource(func)
    return msg


def httplib_request(host, path, params, request='POST', debug=0):
    """request method using provided HTTP request and httplib library"""
    if debug:
        httplib.HTTPConnection.debuglevel = 1
    if not isinstance(params, str):
        params = urllib.parse.urlencode(params, doseq=True)
    if debug:
        print("input parameters", params)
    headers = {"Content-type": "application/x-www-form-urlencoded",
               "Accept": "text/plain"}
    if host.find('https://') != -1:
        host = host.replace('https://', '')
        conn = httplib.HTTPSConnection(host)
    else:
        host = host.replace('http://', '')
        conn = httplib.HTTPConnection(host)
    if request == 'GET':
        conn.request(request, path)
    else:
        conn.request(request, path, params, headers)
    response = conn.getresponse()

    if response.reason != "OK":
        print(response.status, response.reason, response.read())
        res = None
    else:
        res = response.read()
    conn.close()
    return res


def print_data(data, lookup="value"):
    """
    Provides pretty print of json data based on lookup parameter
    by default lookup=value. See couch db response for more info.
    """
    jsondict = json.loads(data)
    PM.print_blue("Total %s documents" % len(jsondict['rows']))
    maxl = 0
    padding = ""
    for row in jsondict['rows']:
        values = row[lookup]
        if isinstance(values, dict):
            if not padding:
                for key in values.keys():
                    if len(key) > maxl:
                        maxl = len(key)
            for key, val in values.items():
                padding = " " * (maxl - len(key))
                print("%s%s: %s" % (padding, PM.msg_blue(key), val))
            print()
        else:
            print(values)


def set_prompt(in1):
    """Define shell prompt"""
    if in1.find('|\#>') != -1:
        in1 = in1.replace('|\#>', '').strip()
    ip = __main__.__dict__['__IP']
    prompt = getattr(ip.outputcache, 'prompt1')
    prompt.p_template = in1 + " |\#> "
    prompt.set_p_str()


def couch_help(self, arg):
    """
    Provide simple help about available commands
    """
    pmgr = PrintManager()
    global magic_list
    msg = "\nAvailable commands:\n"
    for name, func in magic_list:
        msg += "%s\n%s\n" % (pmgr.msg_blue(name), pmgr.msg_green(func.__doc__))
    msg += "List of pre-defined variables to control your interactions "
    msg += "with CouchDB:\n"
    msg += pmgr.msg_green("    URI, DB, DESIGN, DEBUG\n")
    print(msg)


### MAGIC COMMANDS ###
def db_info():
    """
    Provide information about Couch DB. Use DB parameter to setup
    your couch DB name.
    """
    host = URI
    path = '/%s' % DB
    data = httplib_request(host, path, {}, 'GET', DEBUG)
    if not data:
        return data
    return json.loads(data)


def couch_views():
    """
    List registered views in couch db.
    """
    qqq = 'startkey=%22_design%2F%22&endkey=%22_design0%22'
    host = URI
    path = '/%s/_all_docs?%s' % (DB, qqq)
    results = httplib_request(host, path, {}, 'GET', DEBUG)
    designdocs = json.loads(results)
    results = {}
    for item in designdocs['rows']:
        doc = item['key']
        print(PM.msg_blue("design: ") + doc)
        path = '/%s/%s' % (DB, doc)
        res = httplib_request(host, path, {}, 'GET', DEBUG)
        rdict = json.loads(res)
        for view_name, view_dict in rdict['views'].items():
            print(PM.msg_blue("view name: ") + view_name)
            print(PM.msg_blue("map:"))
            print(PM.msg_green(view_dict['map']))
            if 'reduce' in view_dict:
                print(PM.msg_blue("reduce:"))
                print(PM.msg_green(view_dict['reduce']))


def create_view(view_dict):
    """
    Create couch db view. The db and design names are controlled via
    DB and DESIGN shell parameters, respectively.
    Parameters: <view_dict>
    Example of the view:
    {"view_name": {"map" : "function(doc) { if(doc.hash) {emit(1, doc.hash);}}" }}
    """
    # get existing views
    host = URI
    path = '/%s/_design/%s' % (DB, DESIGN)
    data = httplib_request(host, path, {}, 'GET', DEBUG)
    jsondict = json.loads(data)
    for view_name, view_def in view_dict.items():
        jsondict['views'][view_name] = view_def

    # update views
    encoder = JSONEncoder()
    params = encoder.encode(jsondict)
    request = 'PUT'
    debug = DEBUG
    data = httplib_request(host, path, params, request, debug)
    if not data:
        return data
    return json.loads(data)


def delete_view(view_name):
    """
    Delete couch db view. The db and design names are controlled via
    DB and DESIGN shell parameters, respectively.
    Parameters: <view_name>
    """
    # get existing views
    host = URI
    path = '/%s/_design/%s' % (DB, DESIGN)
    data = httplib_request(host, path, {}, 'GET', DEBUG)
    if not data:
        return data
    jsondict = json.loads(data)

    # delete requested view in view dict document
    try:
        del jsondict['views'][view_name]
        # update view dict document in a couch
        encoder = JSONEncoder()
        params = encoder.encode(jsondict)
        request = 'PUT'
        debug = DEBUG
        data = httplib_request(host, path, params, request, debug)
    except:
        traceback.print_exc()


def delete_all_views(design):
    """
    Delete all views in particular design document.
    The db and design names are controlled via
    DB and DESIGN shell parameters, respectively.
    Parameters: <design_name, e.g. dasadmin>
    """
    host = URI
    path = '/%s/_design/%s' % (DB, design)
    data = httplib_request(host, path, {}, 'DELETE', DEBUG)
    if not data:
        return data
    return json.loads(data)


def create_db(db_name):
    """
    Create a new DB in couch.
    Parameters: <db_name>
    """
    host = URI
    path = '/%s' % db_name
    data = httplib_request(host, path, {}, 'PUT', DEBUG)
    if not data:
        return data
    return json.loads(data)


def delete_db(db_name):
    """
    Delete DB in couch.
    Parameters: <db_name>
    """
    host = URI
    path = '/%s' % db_name
    data = httplib_request(host, path, {}, 'DELETE', DEBUG)
    if not data:
        return data
    return json.loads(data)


def get_all_docs(idx=0, limit=0, pretty_print=False):
    """
    Retrieve all documents from CouchDB.
    Parameters: <idx=0> <limit=0> <pretty_print=False>
    """
    host = URI
    path = '/%s/_all_docs' % DB
    kwds = {}
    if idx:
        kwds['skip'] = idx
    if limit:
        kwds['limit'] = limit
    data = httplib_request(host, path, kwds, 'GET', DEBUG)
    if not data:
        return data
    if pretty_print:
        print_data(data, lookup='id')
    else:
        return json.loads(data)


def get_doc(id):
    """
    Retrieve document with given id from CouchDB.
    Parameters: <id, e.g. 1323764f7a6af1b37b72119920cbaa08>
    """
    host = URI
    path = '/%s/%s' % (DB, id)
    kwds = {}
    data = httplib_request(host, path, kwds, 'GET', DEBUG)
    if not data:
        return data
    return json.loads(data)


def load_module(arg):
    """
    Load custom admin module. Name it as <module>_ipython.py and place in
    your PYTHONPATH. Implement a <module>_load() function to load your stuff.
    Parameters: <module>

    Example:
    def mycmd():
        pass # do something here
    def mymodule_load():
        ip = IPython.ipapi.get()
        ip.expose_magic('mycmd', mycmd)
    """
    # try to load custom modules
    stm = "from %s_ipython import %s_load\n" % (arg, arg)
    stm += "%s_load()" % arg
    obj = compile(str(stm), '<string>', 'exec')
    try:
        eval(obj)
        msg = "Loaded %s module. " % arg
        msg += "Use " + PM.msg_blue("%s_help" % arg) + \
               " for concrete module help if it's implemented"
        print(msg)
    except:
        traceback.print_exc()
        pass


# keep magic list as global since it's used in couch_help
magic_list = [
    ('db_info', db_info),
    ('couch_views', couch_views),
    ('create_view', create_view),
    ('delete_view', delete_view),
    ('delete_all_views', delete_all_views),
    ('get_all_docs', get_all_docs),
    ('get_doc', get_doc),
    ('create_db', create_db),
    ('delete_db', delete_db),
    ('load_module', load_module),
]


def main():
    """
    Main function which defint ipython behavior
    """
    pmgr = PrintManager()

    # global IP API
    ip = IPython.ipapi.get()

    o = ip.options
    ip.expose_magic('couch_help', couch_help)
    # load commands and expose them to the shell
    for m in magic_list:
        ip.ex(inspect.getsource(m[1]))

    # load configuration for couch-sh, supply a names of functions to
    # be loaded
    ip.ex(load_config([httplib_request, print_data]))

    # autocall to "full" mode (smart mode is default, I like full mode)
    o.autocall = 2

    # Set dbsh prompt
    o.prompt_in1 = 'couch-sh |\#> '
    o.prompt_in2 = 'couch-sh> '
    o.system_verbose = 0

    # define couch-sh banner
    pyver = sys.version.split('\n')[0]
    ipyver = Release.version
    msg = "Welcome to couch-sh \n[python %s, ipython %s]\n%s\n" \
          % (pyver, ipyver, os.uname()[3])
    msg += "For couch-sh help use "
    msg += pmgr.msg_blue("couch_help")
    msg += ", for python help use help commands\n"
    o.banner = msg
    o.prompts_pad_left = "1"
    # Remove all blank lines in between prompts, like a normal shell.
    o.separate_in = "0"
    o.separate_out = "0"
    o.separate_out2 = "0"


main()



from __future__ import (division, print_function)

from time import time

from WMCore.REST.CherryPyPeriodicTask import CherryPyPeriodicTask
from WMCore.WorkQueue.WorkQueue import globalQueue
from WMCore.WorkQueue.WorkQueueReqMgrInterface import WorkQueueReqMgrInterface


class ReqMgrInteractionTask(CherryPyPeriodicTask):

    def __init__(self, rest, config):
        super(ReqMgrInteractionTask, self).__init__(config)
        self.globalQ = globalQueue(logger=self.logger, **config.queueParams)
        self.reqMgrInt = WorkQueueReqMgrInterface(logger=self.logger, **config.reqMgrConfig)

    def setConcurrentTasks(self, config):
        """
        sets the list of function reference for concurrent tasks
        """
        self.concurrentTasks = [{'func': self.interactWithReqmgr, 'duration': config.interactDuration}]

    def interactWithReqmgr(self, config):
        """
        1. pull new work
        2. add the new element from running-open request
        3. report element status to reqmgr (need to be removed and set as reqmgr task)
        4. record this activity
        """
        tStart = time()
        self.logger.info("Executing WorkQueue/ReqMgr thread")
        self.reqMgrInt(self.globalQ)
        self.logger.info("%s executed in %.3f secs.\n", self.__class__.__name__, time() - tStart)

        return



# -*- coding: utf-8 -*-
"""
_SetState_

MySQL implementation of Locations.SetState

Created on Mon Jun 18 13:16:00 2012

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter


class SetState(DBFormatter):
    sql = """UPDATE wmbs_location
             SET state = (SELECT id FROM wmbs_location_state WHERE name = :state),
             state_time = :state_time
             WHERE site_name = :location
             """

    def execute(self, siteName, state, stateTime, conn=None, transaction=False):
        binds = {"location": siteName, "state": state, "state_time": stateTime}
        self.dbi.processData(self.sql, binds, conn=conn,
                             transaction=transaction)
        return



"""
MySQL implementation of File.Get
"""
from WMCore.Database.DBFormatter import DBFormatter

class GetByID(DBFormatter):
    sql = """SELECT id, lfn, filesize, events, first_event, merged
             FROM wmbs_file_details WHERE id = :fileid"""

    def formatOneDict(self, result):
        """
        _formatOneDict_

        Cast the integer attributes of the file object to integers as
        formatDict() will turn everything into a string.
        """
        formattedResult = DBFormatter.formatDict(self, result)[0]
        formattedResult["id"] = int(formattedResult["id"])
        formattedResult["events"] = int(formattedResult["events"])
        formattedResult["first_event"] = int(formattedResult["first_event"])
        formattedResult["merged"] = bool(int(formattedResult["merged"]))

        if "size" in formattedResult:
            formattedResult["size"] = int(formattedResult["size"])
        else:
            # The size column is named "filesize" in Oracle as size is
            # as reserved word.  We'll handle this here to make things
            # easier in the Oracle version of this object.
            formattedResult["size"] = int(formattedResult["filesize"])
            del formattedResult["filesize"]

        return formattedResult

    def formatBulkDict(self, result):
        """
        _formatBulkDict_

        Formats a whole list of dictionaries
        """

        formattedResult = {}
        listOfDicts     = DBFormatter.formatDict(self, result)

        for entry in listOfDicts:
            tmpDict = {}
            tmpDict["id"]          = int(entry["id"])
            tmpDict["lfn"]         = entry["lfn"]
            tmpDict["events"]      = int(entry["events"])
            tmpDict["first_event"] = int(entry["first_event"])
            if "size" in entry:
                tmpDict["size"]    = int(entry["size"])
            else:
                tmpDict["size"]    = int(entry["filesize"])
                del entry["filesize"]
            formattedResult[tmpDict['id']] = tmpDict

        return formattedResult

    def execute(self, file = None, conn = None, transaction = False):

        #Making some modifications to allow it to load a whole list of files
        #This DAO object should be called directly, not through WMBSFile
        if isinstance(file, list):
            #Then we have a list of the form [fileid, fileid, etc.]
            if len(file) == 0:
                #Ignore empty lists
                return {}
            binds = []
            for id in file:
                binds.append({'fileid': id})

            result = self.dbi.processData(self.sql, binds,
                                          conn = conn, transaction = transaction)
            return self.formatBulkDict(result)
        else:
            #We only have one file ID
            binds = {"fileid": file}
            result = self.dbi.processData(self.sql, binds,
                                          conn = conn, transaction = transaction)
            return self.formatOneDict(result)



#!/usr/bin/env python
# encoding: utf-8
"""
Interface.py

Created by Dave Evans on 2010-07-21.
Copyright (c) 2010 Fermilab. All rights reserved.
"""
from __future__ import print_function



from builtins import str, object

import WMCore.Database.CMSCouch as CMSCouch

class CouchConnectionError(Exception):
    """docstring for CouchConnectionError"""
    def __init__(self, arg):
        super(CouchConnectionError, self).__init__()
        self.arg = arg


class Interface(object):
    def __init__(self, couchUrl, couchDatabase):
        self.cdb_url = couchUrl
        self.cdb_database = couchDatabase
        try:
            self.cdb_server = CMSCouch.CouchServer(self.cdb_url)
            self.couch = self.cdb_server.connectDatabase(self.cdb_database)
        except Exception as ex:
            msg = "Exception instantiating couch services for :\n"
            msg += " url = %s\n database = %s\n" % (self.cdb_url, self.cdb_database)
            msg += " Exception: %s" % str(ex)
            print(msg)
            raise CouchConnectionError(msg)


    def documentsOwned(self, group, user):
        """
        _documentsOwned_

        Get a list of doc IDs that are owned by the group/user pair

        """
        result = self.couch.loadView("GroupUser", 'owner_group_user',
                 {'startkey' :[group, user],
                   'endkey' : [group, user]}, []
                )
        output = [str(x[u'id']) for x in result[u'rows']]
        return output



    def callUpdate(self, update, document, **args ):
        """
        _callUpdate_

        Wrapper primarily for calling/testing update methods on the groupuser couchapp

         * update - name of the update method in the couchapp
         * document - ID of the document to be updated
         * args  - key:value dict of args to be encoded in the PUT request

        """
        updateUri = "/" + self.couch.name + "/_design/GroupUser/_update/"+ update + "/" + document
        argsstr = "?"
        for k, v in list(args.items()):
            argsstr += "%s=%s&" % (k, v)
        updateUri += argsstr
        updateUri= updateUri[:-1]
        self.couch.makeRequest(uri = updateUri, type = "PUT", decode = False)
        return



#!/usr/bin/env python
#
# Originally from myproxy_logon script:
# http://www.mcs.anl.gov/fl/research/accessgrid/myproxy/myproxy_logon.py
# By Tom Uram <turam@mcs.anl.gov> on 2005/08/04
#
# Adapted/improved by
# Diego da Silva Gomes <diego@cern.ch>
# on 2013/04/08 for use by CMSWEB services.
#
# Added SimpleMyProxy class for WMCore.Credential module compatibility.
# Mattia Cinquilli <mcinquil@cern.ch>
# on 2013/04/14 for integration on CRAB services.

from builtins import str
import logging
import re
import socket

from OpenSSL import SSL, crypto

from WMCore.Credential.Credential import Credential

# we don't require password as we only want to allow x509 authentication
CMD_GET = """VERSION=MYPROXYv2
COMMAND=0
USERNAME=%s
PASSPHRASE=
LIFETIME=%d\0"""

CMD_INFO = """VERSION=MYPROXYv2
COMMAND=2
USERNAME=%s
PASSPHRASE=PASSPHRASE
LIFETIME=0\0"""


class MyProxyException(Exception):
    pass


def myproxy_ctx(certfile, keyfile):
    ctx = SSL.Context(SSL.SSLv3_METHOD)
    ctx.use_certificate_chain_file(certfile)
    ctx.use_privatekey_file(keyfile)

    # disable for compatibility with myproxy server (er, globus)
    # globus doesn't handle this case, apparently, and instead
    # chokes in proxy delegation code
    ctx.set_options(SSL.OP_DONT_INSERT_EMPTY_FRAGMENTS)

    return ctx


def create_cert_req(keyType=crypto.TYPE_RSA,
                    bits=1024,
                    messageDigest="md5"):
    """
    Create certificate request.

    Returns: certificate request PEM text, private key PEM text
    """

    # Create certificate request
    req = crypto.X509Req()

    # Generate private key
    pkey = crypto.PKey()
    pkey.generate_key(keyType, bits)

    req.set_pubkey(pkey)
    req.sign(pkey, messageDigest)

    return (crypto.dump_certificate_request(crypto.FILETYPE_ASN1, req),
            crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey))


def deserialize_response(msg):
    m = re.search('RESPONSE=(\d)\n', msg)
    if m:
        resp = int(m.group(1))
    else:
        resp = 1  # set error if response not found

    errors = ", ".join(re.findall('^ERROR=(.*?)\n', msg, re.MULTILINE))
    data = "".join(re.findall('^(?!VERSION|RESPONSE|ERROR)(.*?\n)', msg, re.MULTILINE))
    return resp, errors, data


def deserialize_certs(inp_dat):
    pem_certs = []

    dat = inp_dat

    while dat:

        # find start of cert, get length
        ind = dat.find('\x30\x82')
        if ind < 0:
            break

        length = 256 * ord(dat[ind + 2]) + ord(dat[ind + 3])

        # extract der-format cert, and convert to pem
        c = dat[ind:ind + length + 4]
        x509 = crypto.load_certificate(crypto.FILETYPE_ASN1, c)
        pem_cert = crypto.dump_certificate(crypto.FILETYPE_PEM, x509)
        pem_certs.append(pem_cert)

        # trim cert from data
        dat = dat[ind + length + 4:]

    return pem_certs


def myproxy_client(sslctx, op, username, logger, lifetime=43200, host="myproxy.cern.ch", port=7512):
    """
    Function to info|get a proxy credential from a MyProxy server

    Exceptions: MyProxyException or any of the SSL exceptions
    """
    if op not in ['info', 'get']:
        raise MyProxyException('Wrong operation. Select "info" or "get".')

    logger.debug("debug: connect to myproxy server")
    conn = SSL.Connection(sslctx, socket.socket())
    conn.connect((host, port))

    logger.debug("debug: send globus compat byte")
    conn.write('0')

    logger.debug("debug: send the operation command")
    if op == 'info':
        cmd = CMD_INFO % username
    else:
        cmd = CMD_GET % (username, lifetime)
    conn.write(cmd)

    logger.debug("debug: process server response")
    d = conn.recv(8192)
    logger.debug(d)

    resp, error, data = deserialize_response(d)
    if resp:
        raise MyProxyException(error)
    logger.debug("debug: server response ok")

    if op == 'get':
        # The client will generate a public/private key pair and send a
        # NULL-terminated PKCS#10 certificate request to the server.
        logger.debug("debug: generate and send certificate request")
        certreq, privkey = create_cert_req()
        conn.send(certreq)

        logger.debug("debug: receive the number of certs")
        d = conn.recv(1)
        numcerts = ord(d[0])

        logger.debug("debug: receive %d certs" % numcerts)
        d = conn.recv(8192)

        logger.debug("debug: process server response")
        r = conn.recv(8192)
        resp, error, data = deserialize_response(r)
        if resp:
            raise MyProxyException(error)
        logger.debug("debug: server response ok")

        # deserialize certs from received cert data
        pem_certs = deserialize_certs(d)
        if len(pem_certs) != numcerts:
            raise MyProxyException("%d certs expected, %d received" % (numcerts, len(pem_certs)))
        logger.debug("debug: certs deserialized successfuly")

        # return proxy, the corresponding privkey, and then the rest of cert chain
        data = pem_certs[0] + privkey
        for c in pem_certs[1:]:
            data += c

    return data


class SimpleMyProxy(Credential):
    def __init__(self, args):
        Credential.__init__(self, args)
        self.logger = args['logger'] if 'logger' in args else logging.getLogger(type(self).__name__)

    def checkMyProxy(self, username, certfile=None, keyfile=None, myproxyserver='myproxy.cern.ch', myproxyport=7512):
        """
        Check if a valid myproxy exists and returns related info.
        This is supposed to be executed by the myproxy creator.
        """
        sslctx = myproxy_ctx(certfile if certfile else self.serverCert, keyfile if keyfile else self.serverKey)
        myproxy = myproxy_client(sslctx, 'info', username if username else self.userName, logger=self.logger,
                                 host=myproxyserver, port=myproxyport)
        myproxystatus = {}
        for line in myproxy.split('\n'):
            if line:
                if 'CRED_START_TIME' in line:
                    myproxystatus['start'] = line.split('CRED_START_TIME=')[1]
                elif 'CRED_END_TIME' in line:
                    myproxystatus['end'] = line.split('CRED_END_TIME=')[1]
                elif 'CRED_OWNER' in line:
                    myproxystatus['owner'] = line.split('CRED_OWNER=')[1]
                elif 'CRED_RETRIEVER_TRUSTED' in line:
                    myproxystatus['retriever'] = line.split('CRED_RETRIEVER_TRUSTED=')[1]
        return myproxystatus

    def logonRenewMyProxy(self, username, lifetime=43200, certfile=None, keyfile=None, myproxyserver='myproxy.cern.ch',
                          myproxyport=7512):
        """
        Retrieves a proxy from an existing myproxy server already delegated.
        """
        sslctx = myproxy_ctx(certfile if certfile else self.serverCert, keyfile if keyfile else self.serverKey)
        myproxy = myproxy_client(sslctx, 'get', username if username else self.userName, lifetime=lifetime,
                                 host=myproxyserver, port=myproxyport, logger=self.logger)
        return myproxy


if __name__ == '__main__':
    import sys
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-o", "--op", dest="op",
                        help="Operation: info|get")
    parser.add_argument("-s", "--pshost", dest="host", default="myproxy.cern.ch",
                        help="The hostname of the MyProxy server to contact")
    parser.add_argument("-p", "--psport", dest="port", default=7512,
                        help="The port of the MyProxy server to contact")
    parser.add_argument("-l", "--username", dest="username",
                        help="The username with which the credential is stored on the MyProxy server")
    parser.add_argument("-C", "--certfile", dest="cert",
                        help="Certificate PEM filename to use with myproxy authentication")
    parser.add_argument("-y", "--keyfile", dest="key",
                        help="Certkey PEM filename to use with myproxy authentication")
    parser.add_argument("-t", "--proxy_lifetime", dest="lifetime", default=43200,
                        help="The lifetime validity for the fetched proxy.")
    parser.add_argument("-v", "--verbose", action="store_true", dest="debug", default=False,
                        help="Print debug information to stdout.")
    opt = parser.parse_args()


    def getLogging(debug):
        """Retrieves a logger and set the proper level

        :arg bool debug: it tells if needs a verbose logger
        :return logger: a logger with the appropriate logger level."""
        loglevel = logging.INFO
        if debug:
            loglevel = logging.DEBUG
        logging.basicConfig(level=loglevel)
        logger = logging.getLogger(__name__)
        logger.debug("Logging level initialized to %s.", loglevel)
        return logger


    logger = getLogging(debug=opt.debug)

    # process options
    if not opt.op or opt.op not in ['info', 'get']:
        logger.error("Error: incorrect operation, use --op (see -h for help)")
        sys.exit(1)
    if not opt.username:
        logger.error("Error: username not specified, use --username (see -h for help)")
        sys.exit(1)
    if not opt.cert or not opt.key:
        logger.error("Error: certificate/key not specified, use --certfile and --keyfile (see -h for help)")
        sys.exit(1)
    dbgfunc = lambda x: sys.stdout.write(x + '\n') if opt.debug else 0

    # Do the operation
    try:
        ctx = myproxy_ctx(opt.cert, opt.key)
        ret = myproxy_client(ctx, opt.op, opt.username, lifetime=int(opt.lifetime),
                             host=opt.host, port=int(opt.port), logger=logger)
        sys.stdout.write(ret)
    except Exception as e:
        if opt.debug:
            import traceback

            traceback.print_exc()
        logger.error("Error: %s", str(e))
        sys.exit(1)
